# mypy: disable-error-code="call-overload,misc"
#
# Mypy errors from passing union types directly into tyro.cli() as Type[T]. We would
# benefit from TypeForm[T]: https://github.com/python/mypy/issues/9773
"""Tests for collections containing unions with different fixed nargs.

This tests the functionality added to support types like:
- List[bool| Tuple[int, int]]
- List[Tuple[int, int]| Tuple[int, int, int]]
"""

import dataclasses
from typing import Any, List, Tuple

import pytest
from helptext_utils import get_helptext_with_checks

import tyro


def test_list_bool_or_tuple() -> None:
    """Test list containing union of bool and fixed-size tuple."""
    assert tyro.cli(List[bool | Tuple[int, int]], args="3 4 5 6".split(" ")) == [
        (3, 4),
        (5, 6),
    ]
    assert tyro.cli(
        List[bool | Tuple[int, int]], args="True 3 4 5 6 False".split(" ")
    ) == [
        True,
        (3, 4),
        (5, 6),
        False,
    ]


def test_list_union_different_tuple_sizes() -> None:
    """Test list containing union of tuples with different sizes."""

    def main(
        x: List[Tuple[int, int] | Tuple[int, int, int]],
    ) -> List[Tuple[int, int] | Tuple[int, int, int]]:
        return x

    # This is now supported! Test that it works correctly.
    assert tyro.cli(main, args=["--x", "1", "2", "3", "4"]) == [(1, 2), (3, 4)]
    assert tyro.cli(main, args=["--x", "1", "2", "3", "4", "5", "6"]) == [
        (1, 2),
        (3, 4),
        (5, 6),
    ]

    # With backtracking, "1 2 3" now works!
    assert tyro.cli(main, args=["--x", "1", "2", "3"]) == [(1, 2, 3)]


def test_list_union_different_tuple_sizes_direct() -> None:
    """Test directly passing the type annotation to tyro.cli."""
    X = List[Tuple[int, int] | Tuple[int, int, int]]
    # This is now supported!
    assert tyro.cli(X, args=["1", "2", "3", "4"]) == [(1, 2), (3, 4)]
    assert tyro.cli(X, args=["1", "2", "3", "4", "5", "6"]) == [(1, 2), (3, 4), (5, 6)]


def test_union_over_collections_from_test_collections() -> None:
    """Test from test_collections.py that now works."""

    def main(a: Tuple[int, int] | Tuple[int, int, int]) -> Tuple[int, ...]:
        return a

    assert tyro.cli(main, args=["--a", "5", "5"]) == (5, 5)
    assert tyro.cli(main, args=["--a", "5", "5", "2"]) == (5, 5, 2)
    with pytest.raises(SystemExit):
        tyro.cli(main, args=["--a", "5", "5", "2", "1"])
    with pytest.raises(SystemExit):
        tyro.cli(main, args=["--a"])


def test_list_union_with_str_and_tuple() -> None:
    """Test list containing union of string and tuple."""

    def main(x: List[str | Tuple[int, int]]) -> List[str | Tuple[int, int]]:
        return x

    # When str is in a union with tuple, str will match because both have fixed nargs
    # and the parsing tries options from left to right.
    assert tyro.cli(main, args=["--x", "hello", "1", "2", "world"]) == [
        "hello",
        "1",
        "2",
        "world",
    ]
    assert tyro.cli(main, args=["--x", "1", "2", "3", "4"]) == ["1", "2", "3", "4"]

    # Even with tuple first, strings that can be parsed as strings will be.
    def main2(
        x: List[Tuple[int, int] | str],
    ) -> List[Tuple[int, int] | str]:
        return x

    # When all args can be parsed as strings, they will be.
    assert tyro.cli(main2, args=["--x", "1", "2", "hello", "3", "4"]) == [
        "1",
        "2",
        "hello",
        "3",
        "4",
    ]


def test_list_union_triple_tuple_sizes() -> None:
    """Test list containing union of three different tuple sizes."""

    def main(
        x: List[Tuple[int] | Tuple[int, int] | Tuple[int, int, int]],
    ) -> List[Tuple[int] | Tuple[int, int] | Tuple[int, int, int]]:
        return x

    # Should parse greedily, trying single int first.
    assert tyro.cli(main, args=["--x", "1", "2", "3"]) == [(1,), (2,), (3,)]
    assert tyro.cli(main, args=["--x", "1", "2", "3", "4", "5", "6"]) == [
        (1,),
        (2,),
        (3,),
        (4,),
        (5,),
        (6,),
    ]


def test_dataclass_with_list_union_tuples() -> None:
    """Test dataclass containing list of union tuples."""

    @dataclasses.dataclass
    class Config:
        points: List[Tuple[int, int] | Tuple[int, int, int]]
        names: List[str]

    result = tyro.cli(
        Config, args=["--points", "1", "2", "3", "4", "--names", "a", "b"]
    )
    assert result.points == [(1, 2), (3, 4)]
    assert result.names == ["a", "b"]


def test_nested_list_union() -> None:
    def main(x: List[List[int | Tuple[int, str]]]) -> Any:
        return x

    assert tyro.cli(main, args=["--x", "1", "2", "three", "4"]) == [
        [1, (2, "three"), 4]
    ]


def test_help_message_union_tuples() -> None:
    """Test that help messages show the union options correctly."""

    def main(x: List[Tuple[int, int] | Tuple[int, int, int]]) -> None:
        pass

    # Capture help output.
    with pytest.raises(SystemExit):
        tyro.cli(main, args=["--help"])
    # The metavar should show both options: {INT INT}|{INT INT INT}.


def test_edge_cases() -> None:
    """Test edge cases for the new functionality."""
    # Empty list.
    assert tyro.cli(List[bool | Tuple[int, int]], args=[]) == []

    # Single bool.
    assert tyro.cli(List[bool | Tuple[int, int]], args=["True"]) == [True]
    assert tyro.cli(List[bool | Tuple[int, int]], args=["False"]) == [False]

    # Single tuple.
    assert tyro.cli(List[bool | Tuple[int, int]], args=["1", "2"]) == [(1, 2)]

    # Mixed types in specific order.
    assert tyro.cli(
        List[bool | Tuple[int, int]], args=["1", "2", "True", "3", "4", "False"]
    ) == [(1, 2), True, (3, 4), False]


def test_backtracking_parser() -> None:
    """Test cases that now work with backtracking parser."""

    # Case 1: Input "1 2 3 4 5" with Tuple[int, int]| Tuple[int, int, int].
    # Backtracking finds (1,2), (3,4,5).
    result = tyro.cli(
        List[Tuple[int, int] | Tuple[int, int, int]],
        args=["1", "2", "3", "4", "5"],
    )
    assert result == [(1, 2), (3, 4, 5)]

    # Case 2: Input "1 2 3" with Tuple[int, int]| Tuple[int, int, int].
    # Backtracking finds (1,2,3).
    result = tyro.cli(
        List[Tuple[int, int] | Tuple[int, int, int]], args=["1", "2", "3"]
    )
    assert result == [(1, 2, 3)]

    # Case 3: nargs are sorted, so (3,2) becomes (2,3).
    result = tyro.cli(
        List[Tuple[int, int, int] | Tuple[int, int]],
        args=["1", "2", "3", "4", "5"],
    )
    # With sorted nargs (2,3), backtracking finds (1,2), (3,4,5).
    assert result == [(1, 2), (3, 4, 5)]


def test_long_list_recursion_limit() -> None:
    """Test that we can handle very long lists without hitting recursion limit."""

    def main(l: List[str]) -> List[str]:
        return l

    # Test with 1000 elements - this would previously cause RecursionError.
    args = ["--l"] + ["a"] * 1000
    result = tyro.cli(main, args=args)
    assert result == ["a"] * 1000

    # Test with 2000 elements for good measure.
    args = ["--l"] + ["b"] * 2000
    result = tyro.cli(main, args=args)
    assert result == ["b"] * 2000


def test_truly_unparseable() -> None:
    """Test cases that are truly unparseable even with backtracking."""

    # Case 1: Input "1 2 3 4" with Tuple[int, int, int]| Tuple[int, int, int, int, int].
    # No valid parse exists.
    with pytest.raises(SystemExit):
        tyro.cli(
            List[Tuple[int, int, int] | Tuple[int, int, int, int, int]],
            args=["1", "2", "3", "4"],
        )

    # Case 2: Single element that can't match any option.
    with pytest.raises(SystemExit):
        tyro.cli(List[bool | Tuple[int, int]], args=["1"])


def test_greedy_parsing_success_1() -> None:
    """Test cases where greedy parsing happens to work."""

    # Works because we try smallest first.
    result = tyro.cli(
        List[Tuple[int] | Tuple[int, int] | Tuple[int, int, int]],
        args=["1", "2", "3", "4", "5", "6"],
    )
    assert result == [(1,), (2,), (3,), (4,), (5,), (6,)]


def test_greedy_parsing_success_2() -> None:
    """Test cases where greedy parsing happens to work."""
    # Works because exact multiples.
    result = tyro.cli(
        List[Tuple[int, int] | Tuple[int, int, int]], args=["1", "2", "3", "4"]
    )
    assert result == [(1, 2), (3, 4)]


def test_nested_union_in_collection() -> None:
    """Test nested union in collection."""

    # Works because exact multiples.
    result = tyro.cli(
        List[Tuple[int, int] | Tuple[int, int, int] | bool],
        args=["1", "2", "3", "4"],
    )
    assert result == [(1, 2), (3, 4)]


def f1(x: List[Tuple[int, int] | Tuple[int, int, int]]) -> None:
    """Main function.

    Args:
        x: List of 2D or 3D points.
    """
    pass


def test_helptext_list_union_tuples() -> None:
    """Test that help messages show the union options correctly."""

    helptext = get_helptext_with_checks(f1)
    assert "--x" in helptext
    # Check that the metavar shows both tuple options.
    assert "{INT INT}|{INT INT INT}" in helptext
    assert "List of 2D or 3D points." in helptext


@dataclasses.dataclass
class Config1:
    values: List[bool | Tuple[int, int]]
    """A list of boolean flags or coordinate pairs."""


def test_helptext_list_bool_or_tuple() -> None:
    """Test helptext for list containing union of bool and tuple."""

    helptext = get_helptext_with_checks(Config1)
    assert "--values" in helptext
    assert "{True,False}|{INT INT}" in helptext
    assert "A list of boolean flags or coordinate pairs." in helptext


@dataclasses.dataclass
class Config2:
    points: List[Tuple[int, int] | Tuple[int, int, int]]
    """List of 2D or 3D points."""
    names: List[str]
    """List of point names."""


def test_helptext_dataclass_with_union_list() -> None:
    """Test helptext for dataclass containing list of union tuples."""

    helptext = get_helptext_with_checks(Config2)
    assert "--points" in helptext
    assert "{INT INT}|{INT INT INT}" in helptext
    assert "List of 2D or 3D points. (required)" in helptext
    assert "--names" in helptext
    assert "List of point names. (required)" in helptext


@dataclasses.dataclass
class Config3:
    points: List[Tuple[int, int] | Tuple[int, int, int]] = dataclasses.field(
        default_factory=lambda: [(1, 2), (3, 4, 5)]
    )
    flags: List[bool | Tuple[int, int]] = dataclasses.field(
        default_factory=lambda: [True, (10, 20), False]
    )


def test_list_union_with_defaults() -> None:
    """Test list of unions with default values."""

    # Test with defaults.
    result = tyro.cli(Config3, args=[])
    assert result.points == [(1, 2), (3, 4, 5)]
    assert result.flags == [True, (10, 20), False]

    # Test overriding defaults.
    result = tyro.cli(Config3, args=["--points", "5", "6", "7", "--flags", "False"])
    assert result.points == [(5, 6, 7)]
    assert result.flags == [False]

    # Test partial override.
    result = tyro.cli(Config3, args=["--points", "1", "1", "2", "2"])
    assert result.points == [(1, 1), (2, 2)]
    assert result.flags == [True, (10, 20), False]  # Default preserved.


def test_list_union_empty_default() -> None:
    """Test list of unions with empty default."""

    @dataclasses.dataclass
    class Config:
        points: List[Tuple[int, int] | Tuple[int, int, int]] = dataclasses.field(
            default_factory=list
        )

    # Test with empty default.
    result = tyro.cli(Config, args=[])
    assert result.points == []

    # Test adding values.
    result = tyro.cli(Config, args=["--points", "1", "2", "3", "4", "5"])
    assert result.points == [(1, 2), (3, 4, 5)]


def test_function_with_default_list_union() -> None:
    """Test function with default list of unions."""

    def main(
        coords: List[Tuple[int, int] | Tuple[int, int, int]] = [
            (0, 0),
            (1, 1, 1),
        ],
    ) -> List[Tuple[int, int] | Tuple[int, int, int]]:
        return coords

    # Test with default.
    assert tyro.cli(main, args=[]) == [(0, 0), (1, 1, 1)]

    # Test override.
    assert tyro.cli(main, args=["--coords", "2", "3", "4", "5", "6"]) == [
        (2, 3),
        (4, 5, 6),
    ]


@dataclasses.dataclass
class Config4:
    points: List[Tuple[int, int] | Tuple[int, int, int]] = dataclasses.field(
        default_factory=lambda: [(1, 2), (3, 4, 5)]
    )
    """List of 2D or 3D points."""


def test_helptext_with_defaults() -> None:
    """Test helptext shows default values correctly."""

    helptext = get_helptext_with_checks(Config4)
    assert "--points" in helptext
    assert "{INT INT}|{INT INT INT}" in helptext
    assert "List of 2D or 3D points." in helptext
    # Should show it has a default value.
    assert "(default:" in helptext
    # Default values are shown as space-separated integers.
    assert "1 2 3 4 5" in helptext


def test_mixed_type_union_defaults() -> None:
    """Test mixed type unions with defaults."""

    @dataclasses.dataclass
    class Config:
        values: List[str | Tuple[int, int]] = dataclasses.field(
            default_factory=lambda: ["hello", (1, 2), "world"]
        )

    # Test with default.
    result = tyro.cli(Config, args=[])
    assert result.values == ["hello", (1, 2), "world"]

    # Since str matches greedily, all inputs become strings.
    result = tyro.cli(Config, args=["--values", "5", "6", "test"])
    assert result.values == ["5", "6", "test"]


def test_triple_union_with_defaults() -> None:
    """Test triple union with default values."""

    def main(
        values: List[Tuple[int] | Tuple[int, int] | Tuple[int, int, int]] = [
            (1,),
            (2, 3),
            (4, 5, 6),
        ],
    ) -> List[Tuple[int] | Tuple[int, int] | Tuple[int, int, int]]:
        return values

    # Test with default.
    assert tyro.cli(main, args=[]) == [(1,), (2, 3), (4, 5, 6)]

    # Test override - greedy parsing means all become single tuples.
    assert tyro.cli(main, args=["--values", "7", "8", "9"]) == [(7,), (8,), (9,)]


def test_tuple_union_direct() -> None:
    """Test direct tuple union: Tuple[int, int]| Tuple[int, int, int]."""
    # Test the direct type annotation as requested.
    assert tyro.cli(Tuple[int, int] | Tuple[int, int, int], args=["5", "5"]) == (
        5,
        5,
    )
    assert tyro.cli(Tuple[int, int] | Tuple[int, int, int], args=["5", "5", "2"]) == (
        5,
        5,
        2,
    )

    # Test that invalid argument counts fail.
    with pytest.raises(SystemExit):
        tyro.cli(Tuple[int, int] | Tuple[int, int, int], args=["5", "5", "2", "1"])
    with pytest.raises(SystemExit):
        tyro.cli(Tuple[int, int] | Tuple[int, int, int], args=["5"])
    with pytest.raises(SystemExit):
        tyro.cli(Tuple[int, int] | Tuple[int, int, int], args=[])



"""Bash completion generation for tyro using embedded Python logic."""

from __future__ import annotations

from ... import _parsers
from ._base import CompletionGenerator
from ._python_completer import get_embedded_code
from ._spec import build_completion_spec


class TyroBashCompletionGenerator(CompletionGenerator):
    """Generates bash completion scripts with embedded Python logic."""

    def generate(
        self,
        parser_spec: _parsers.ParserSpecification,
        prog: str,
        root_prefix: str,
    ) -> str:
        """Generate a bash completion script.

        Args:
            parser_spec: Parser specification to generate completion for.
            prog: Program name.
            root_prefix: Prefix for completion function names.

        Returns:
            Bash completion script as a string.
        """
        # Build completion spec.
        spec = build_completion_spec(parser_spec, prog)
        spec_repr = repr(spec)

        # Get embedded Python code.
        python_code = get_embedded_code()

        # Generate the bash completion script.
        script = f"""# AUTOMATICALLY GENERATED by tyro
# Bash completion for {prog}

# Helper function to trim colon-containing prefixes from completions.
# This is necessary because bash treats ':' as a word separator.
# When user types "optimizer:ad", bash splits it as ["optimizer", ":", "ad"].
# We return "optimizer:adam", but bash thinks current word is just "ad",
# so it would append and create "optimizer:optimizer:adam" without this trimming.
__tyro_{root_prefix}_ltrim_colon_completions() {{
  local cur=$1
  # Only trim if the current word contains a colon.
  if [[ "$cur" == *:* ]]; then
    # Remove everything up to and including the last colon from each completion.
    local prefix=${{cur%${{cur##*:}}}}
    local i
    for i in "${{!COMPREPLY[@]}}"; do
      COMPREPLY[$i]=${{COMPREPLY[$i]#"$prefix"}}
    done
  fi
}}

# Main completion function.
_{root_prefix}() {{
  local cur="${{COMP_WORDS[COMP_CWORD]}}"

  # Reconstruct cur if it was split by bash on colons.
  # When typing "optimizer:ad", bash splits it as ["optimizer", ":", "ad"].
  # We need to reconstruct it as "optimizer:ad" for proper completion.
  # However, we must stop at word boundaries (spaces between complete args).
  local i=$COMP_CWORD
  while [[ $i -gt 0 ]]; do
    local prev="${{COMP_WORDS[i-1]}}"
    if [[ "$prev" == ":" ]]; then
      # Previous word is a colon, prepend it and continue.
      cur=":$cur"
      i=$((i-1))
    elif [[ "$prev" != "${{COMP_WORDS[0]}}" && "$prev" != -* ]]; then
      # Check if we're at a word boundary (complete colon-separated value).
      # If the word TWO positions back is a colon, then prev is a complete value.
      # Example: ["dataset", ":", "image-net", "optimizer", ":", "ad"]
      #                                         ^---------^
      #           When at "ad", prev="optimizer", two-back=":"
      #           But optimizer is a NEW argument, not part of image-net.
      if [[ $i -ge 3 && "${{COMP_WORDS[i-2]}}" == ":" ]]; then
        # Two words back is a colon, so prev completed a colon-separated value.
        # Don't merge prev - it's a separate argument.
        break
      fi
      # Previous word is part of the current colon-separated value, prepend it.
      cur="$prev$cur"
      i=$((i-1))
    else
      # Hit program name or an option, stop.
      break
    fi
  done

  # Get completions from Python.
  local python_cmd
  if command -v python3 >/dev/null 2>&1; then
    python_cmd="python3"
  elif command -v python >/dev/null 2>&1; then
    python_cmd="python"
  else
    COMPREPLY=()
    return 1
  fi

  # Pass raw COMP_WORDS array to Python for processing.
  # Python will handle word reconstruction (e.g., merging "dataset" ":" "mnist" -> "dataset:mnist").
  local completions
  completions=$("$python_cmd" - "${{COMP_WORDS[@]}}" "$COMP_CWORD" << 'PYTHON_EOF'
# Hardcoded completion spec.
COMPLETION_SPEC = {spec_repr}

{python_code}
PYTHON_EOF
)

  # Check for special path completion marker.
  if [[ "$completions" == "__TYRO_COMPLETE_FILES__" ]]; then
    COMPREPLY=($(compgen -f -- "$cur"))
    return 0
  fi

  # Parse completions and build COMPREPLY array.
  local IFS=$'\\n'
  local -a matches
  while IFS= read -r line; do
    # Extract the completion part (before the tab).
    local completion="${{line%%$'\\t'*}}"
    matches+=("$completion")
  done <<< "$completions"

  COMPREPLY=("${{matches[@]}}")

  # Trim colon-containing prefixes to handle bash's word splitting on colons.
  __tyro_{root_prefix}_ltrim_colon_completions "$cur"

  return 0
}}

# Register completion for the full program name.
complete -F _{root_prefix} {prog}

# Also register for basename to handle ./script invocations.
_prog_basename=$(basename "{prog}")
if [ "$_prog_basename" != "{prog}" ]; then
  complete -F _{root_prefix} "$_prog_basename"
fi
"""
        return script



"""Interface for generating `argparse.ArgumentParser()` definitions from callables."""

from __future__ import annotations

import dataclasses
import numbers
import warnings
from typing import Any, Callable, Dict, List, Set, Tuple, Type, TypeVar, Union, cast

from typing_extensions import Annotated, get_args, get_origin

from tyro.constructors._registry import ConstructorRegistry
from tyro.constructors._struct_spec import (
    InvalidDefaultInstanceError,
    UnsupportedStructTypeMessage,
)

from . import (
    _arguments,
    _docstrings,
    _fields,
    _resolver,
    _singleton,
    _strings,
    _subcommand_matching,
)
from . import _fmtlib as fmt
from ._typing import TypeForm
from ._typing_compat import is_typing_union
from .conf import _confstruct, _markers
from .constructors._primitive_spec import (
    PrimitiveConstructorSpec,
    UnsupportedTypeAnnotationError,
)

T = TypeVar("T")


@dataclasses.dataclass()
class LazyParserSpecification:
    """Lazy wrapper that defers full ParserSpecification creation until needed.

    Stores lightweight metadata (description) for fast help text generation,
    while deferring expensive parser construction until actually needed.
    """

    # Lightweight field needed for tyro help formatting.
    description: str

    # Factory for creating the full parser when needed.
    _factory: Callable[[], ParserSpecification]
    _cached: ParserSpecification | None = dataclasses.field(default=None, init=False)

    def evaluate(self) -> ParserSpecification:
        """Get the full ParserSpecification, creating it if needed."""
        if self._cached is None:
            self._cached = self._factory()
        return self._cached


@dataclasses.dataclass()
class ArgWithContext:
    arg: _arguments.ArgumentDefinition
    source_parser: ParserSpecification
    """ParserSpecification that directly contains this argument."""
    local_root_parser: ParserSpecification
    """Furthest ancestor of `source_parser` within the same (sub)command."""


@dataclasses.dataclass(frozen=True)
class ParserSpecification:
    """Each parser contains a list of arguments and optionally some subparsers."""

    f: Callable
    markers: Set[_markers._Marker]
    description: str
    args: List[_arguments.ArgumentDefinition]
    field_list: List[_fields.FieldDefinition]
    child_from_prefix: Dict[str, ParserSpecification]
    helptext_from_intern_prefixed_field_name: Dict[
        str, str | Callable[[], str | None] | None
    ]

    # Subparser groups that are direct children of this parser. The actual tree
    # structure for argparse is built on-demand in apply().
    subparsers_from_intern_prefix: Dict[str, SubparsersSpecification]
    intern_prefix: str
    extern_prefix: str
    has_required_args: bool
    subparser_parent: ParserSpecification | None
    prog_suffix: str

    @staticmethod
    def from_callable_or_type(
        f: Callable[..., T],
        markers: Set[_markers.Marker],
        description: str | Callable[[], str | None] | None,
        parent_classes: Set[Type[Any]],
        default_instance: Union[
            T, _singleton.PropagatingMissingType, _singleton.NonpropagatingMissingType
        ],
        intern_prefix: str,
        extern_prefix: str,
        subcommand_prefix: str,
        support_single_arg_types: bool,
        prog_suffix: str,
    ) -> ParserSpecification:
        """Create a parser definition from a callable or type."""

        # Consolidate subcommand types.
        f_unwrapped, new_markers = _resolver.unwrap_annotated(f, _markers._Marker)
        markers = markers | set(new_markers)

        # Cycle detection.
        #
        # - `parent` here refers to in the nesting hierarchy, not the superclass.
        # - We threshold by `max_nesting_depth` to suppress false positives,
        #   for example from custom constructors that behave differently
        #   depending the default value. (example: ml_collections.ConfigDict)
        max_nesting_depth = 128
        if (
            f in parent_classes
            and f is not dict
            and intern_prefix.count(".") > max_nesting_depth
        ):
            raise UnsupportedTypeAnnotationError(
                (
                    fmt.text(
                        "Found a cyclic dependency with type ",
                        fmt.text["cyan"](str(f)),
                    ),
                )
            )

        # TODO: we are abusing the (minor) distinctions between types, classes, and
        # callables throughout the code. This is mostly for legacy reasons, could be
        # cleaned up.
        parent_classes = parent_classes | {cast(Type, f)}

        # Wrap our type with a dummy dataclass if it can't be treated as a
        # nested type. For example: passing in f=int will result in a dataclass
        # with a single field typed as int.
        #
        # Why don't we always use a dummy dataclass?
        # => Docstrings for inner structs are currently lost when we nest struct types.
        from . import _calling

        # Resolve the type of `f`, generate a field list.
        # Try once first to avoid calling field_list_from_type_or_callable twice.
        f_for_field_list = f
        default_instance_for_field_list = default_instance

        # Check if we need DummyWrapper by trying to get fields first.
        with _fields.FieldDefinition.marker_context(tuple(markers)):
            out = _fields.field_list_from_type_or_callable(
                f=f_for_field_list,
                default_instance=default_instance_for_field_list,
                support_single_arg_types=support_single_arg_types,
                in_union_context=False,
            )

            # If not a struct type and not None, wrap in DummyWrapper and try again.
            if isinstance(
                out, UnsupportedStructTypeMessage
            ) and f_unwrapped is not type(None):
                try:
                    f_for_field_list = _calling.DummyWrapper[f]  # type: ignore
                    default_instance_for_field_list = _calling.DummyWrapper(
                        default_instance
                    )  # type: ignore
                    out = _fields.field_list_from_type_or_callable(
                        f=f_for_field_list,
                        default_instance=default_instance_for_field_list,
                        support_single_arg_types=support_single_arg_types,
                        in_union_context=False,
                    )
                except TypeError as e:  # pragma: no cover
                    # In Python 3.8, DummyWrapper[f] raises TypeError if f is not a valid type.
                    # (e.g., "Parameters to generic types must be types. Got 5.")
                    raise UnsupportedTypeAnnotationError(
                        (
                            fmt.text(
                                "Expected a type, class, or callable, but got ",
                                fmt.text["cyan"](repr(f)),
                                ".",
                            ),
                        )
                    ) from e

            assert not isinstance(out, UnsupportedStructTypeMessage), out.message
            assert not isinstance(out, InvalidDefaultInstanceError), "\n".join(
                repr(fmt.rows(*out.message))
            )
            f, field_list = out

        has_required_args = False
        args: list[_arguments.ArgumentDefinition] = []
        helptext_from_intern_prefixed_field_name: Dict[
            str, str | Callable[[], str | None] | None
        ] = {}

        child_from_prefix: Dict[str, ParserSpecification] = {}

        subparsers_from_prefix = {}

        for field in field_list:
            field_out = handle_field(
                field,
                parent_classes=parent_classes,
                intern_prefix=intern_prefix,
                extern_prefix=extern_prefix,
                subcommand_prefix=subcommand_prefix,
                prog_suffix=prog_suffix,
            )
            if isinstance(field_out, _arguments.ArgumentDefinition):
                # Handle single arguments.
                args.append(field_out)
                if field_out.lowered.required:
                    has_required_args = True
            elif isinstance(field_out, SubparsersSpecification):
                # Handle subparsers.
                subparsers_from_prefix[field_out.intern_prefix] = field_out
            elif isinstance(field_out, ParserSpecification):
                # Handle nested parsers.
                nested_parser = field_out
                child_from_prefix[field_out.intern_prefix] = nested_parser

                # Flatten subparsers from nested parser into current parser.
                # This handles the case where a field's type has subcommands that need
                # to be accessible at the parent level.
                for (
                    prefix,
                    subparser_spec,
                ) in nested_parser.subparsers_from_intern_prefix.items():
                    subparsers_from_prefix[prefix] = subparser_spec

                if nested_parser.has_required_args:
                    has_required_args = True

                # Helptext for this field; used as description for grouping arguments.
                class_field_name = _strings.make_field_name(
                    [intern_prefix, field.intern_name]
                )
                if field.helptext is not None:
                    # Keep lazy - don't evaluate yet.
                    helptext_from_intern_prefixed_field_name[class_field_name] = (
                        field.helptext
                    )
                else:
                    helptext_from_intern_prefixed_field_name[class_field_name] = (
                        _docstrings.get_callable_description(nested_parser.f)
                    )

                # If arguments are in an optional group, it indicates that the default_instance
                # will be used if none of the arguments are passed in.
                if (
                    len(nested_parser.args) >= 1
                    and _markers._OPTIONAL_GROUP in nested_parser.args[0].field.markers
                ):
                    current_helptext = helptext_from_intern_prefixed_field_name[
                        class_field_name
                    ]
                    # Evaluate lazy helptext before concatenating.
                    if callable(current_helptext):
                        current_helptext = current_helptext()
                    helptext_from_intern_prefixed_field_name[class_field_name] = (
                        ("" if current_helptext is None else current_helptext + "\n\n")
                        + "Default: "
                        + str(field.default)
                    )

        # Evaluate lazy description if callable.
        desc = (
            description
            if description is not None
            else _docstrings.get_callable_description(f)
        )
        if callable(desc):
            desc = desc()
        # If still None after evaluation, use empty string.
        if desc is None:
            desc = ""

        parser_spec = ParserSpecification(
            f=f,
            markers=markers,
            description=_strings.remove_single_line_breaks(desc),
            args=args,
            field_list=field_list,
            child_from_prefix=child_from_prefix,
            helptext_from_intern_prefixed_field_name=helptext_from_intern_prefixed_field_name,
            subparsers_from_intern_prefix=subparsers_from_prefix,
            intern_prefix=intern_prefix,
            extern_prefix=extern_prefix,
            has_required_args=has_required_args,
            subparser_parent=None,
            prog_suffix=prog_suffix,
        )

        return parser_spec

    def get_args_including_children(
        self,
        local_root: ParserSpecification | None = None,
    ) -> list[ArgWithContext]:
        """Get all arguments in this parser and its children.

        Does not include arguments in subparsers.
        """
        if local_root is None:
            local_root = self
        args = [ArgWithContext(arg, self, local_root) for arg in self.args]
        for child in self.child_from_prefix.values():
            args.extend(child.get_args_including_children(local_root))
        return args


def handle_field(
    field: _fields.FieldDefinition,
    parent_classes: Set[Type[Any]],
    intern_prefix: str,
    extern_prefix: str,
    subcommand_prefix: str,
    prog_suffix: str,
) -> Union[
    _arguments.ArgumentDefinition,
    ParserSpecification,
    SubparsersSpecification,
]:
    """Determine what to do with a single field definition."""

    # Check that the default value matches the final resolved type.
    # There's some similar Union-specific logic for this in narrow_union_type(). We
    # may be able to consolidate this.
    if (
        not _resolver.is_instance(field.type_stripped, field.default)
        # If a custom constructor is set, static_type may not be
        # matched to the annotated type.
        and field.argconf.constructor_factory is None
        and not _singleton.is_sentinel(field.default)
        # The numeric tower in Python is wacky. This logic is non-critical, so
        # we'll just skip it (+the complexity) for numbers.
        and not isinstance(field.default, numbers.Number)
    ):
        # If the default value doesn't match the resolved type, we expand the
        # type. This is inspired by https://github.com/brentyi/tyro/issues/88.
        field_name = _strings.make_field_name([extern_prefix, field.extern_name])
        message = (
            f"The field `{field_name}` is annotated with type `{field.type}`, "
            f"but the default value `{field.default}` has type `{type(field.default)}`. "
            f"We'll try to handle this gracefully, but it may cause unexpected behavior."
        )
        warnings.warn(message)
        field = field.with_new_type_stripped(
            Union[field.type_stripped, type(field.default)]  # type: ignore
        )

    # Force primitive if (1) the field is annotated with a primitive constructor spec, or (2) if
    # a custom primitive exists for the type.
    force_primitive = (
        len(_resolver.unwrap_annotated(field.type, PrimitiveConstructorSpec)[1]) > 0
    ) or ConstructorRegistry._is_primitive_type(
        field.type, field.markers, nondefault_only=True
    )

    if not force_primitive:
        # (1) Handle Unions over callables; these result in subparsers.
        if _markers.Suppress not in field.markers:
            subparsers_attempt = SubparsersSpecification.from_field(
                field,
                parent_classes=parent_classes,
                intern_prefix=_strings.make_field_name(
                    [intern_prefix, field.intern_name]
                ),
                extern_prefix=_strings.make_field_name(
                    [extern_prefix, field.extern_name]
                ),
                prog_suffix=prog_suffix,
            )
            if subparsers_attempt is not None:
                return subparsers_attempt

        # (2) Handle nested callables.
        if _fields.is_struct_type(field.type, field.default, in_union_context=False):
            # Keep description lazy - don't evaluate yet.
            return ParserSpecification.from_callable_or_type(
                field.type_stripped,
                markers=field.markers,
                description=field.helptext,
                parent_classes=parent_classes,
                default_instance=field.default,
                intern_prefix=_strings.make_field_name(
                    [intern_prefix, field.intern_name]
                ),
                extern_prefix=(
                    _strings.make_field_name([extern_prefix, field.extern_name])
                    if field.argconf.prefix_name in (True, None)
                    else field.extern_name
                ),
                subcommand_prefix=subcommand_prefix,
                support_single_arg_types=False,
                prog_suffix=prog_suffix,
            )

    # (3) Handle primitive or fixed types. These produce a single argument!
    arg = _arguments.ArgumentDefinition(
        intern_prefix=intern_prefix,
        extern_prefix=extern_prefix,
        subcommand_prefix=subcommand_prefix,
        field=field,
    )

    # Validate that Fixed/Suppress fields have defaults.
    if (
        _markers.Fixed in field.markers or _markers.Suppress in field.markers
    ) and _singleton.is_missing(field.default):
        raise UnsupportedTypeAnnotationError(
            (
                fmt.text(
                    "Field ",
                    fmt.text["magenta", "bold"](field.intern_name),
                    " is marked as Fixed or Suppress but is missing a default value",
                ),
            )
        )

    return arg


@dataclasses.dataclass(frozen=True)
class SubparsersSpecification:
    """Structure for defining subparsers. Each subparser is a parser with a name."""

    description: str | Callable[[], str | None] | None
    parser_from_name: Dict[str, LazyParserSpecification]
    default_name: str | None
    default_parser: ParserSpecification | None
    intern_prefix: str
    extern_prefix: str
    required: bool
    default_instance: Any
    options: Tuple[Union[TypeForm[Any], Callable], ...]
    prog_suffix: str

    @staticmethod
    def from_field(
        field: _fields.FieldDefinition,
        parent_classes: Set[Type[Any]],
        intern_prefix: str,
        extern_prefix: str,
        prog_suffix: str,
    ) -> SubparsersSpecification | ParserSpecification | None:
        """From a field: return either a subparser specification, a parser
        specification for subcommands when `tyro.conf.AvoidSubcommands` is used
        and a default is set, or `None` if the field does not create a
        subparser."""
        # Union of classes should create subparsers.
        typ = _resolver.unwrap_annotated(field.type_stripped)
        if not is_typing_union(get_origin(typ)):
            return None

        # We don't use sets here to retain order of subcommands.
        options: List[Union[type, Callable]]
        options = [typ for typ in get_args(typ)]

        # If specified, swap types using tyro.conf.subcommand(constructor=...).
        found_subcommand_conf = False
        for i, option in enumerate(options):
            _, found_subcommand_configs = _resolver.unwrap_annotated(
                option, _confstruct._SubcommandConfig
            )
            if (
                len(found_subcommand_configs) > 0
                and found_subcommand_configs[0].constructor_factory is not None
            ):
                found_subcommand_conf = True
                options[i] = Annotated[  # type: ignore
                    (
                        found_subcommand_configs[0].constructor_factory(),
                        *_resolver.unwrap_annotated(option, "all")[1],
                    )
                ]

        # Exit if we don't contain any struct types.
        def recursive_contains_struct_type(options: list[Any]) -> bool:
            for o in options:
                if _fields.is_struct_type(
                    o, _singleton.MISSING_NONPROP, in_union_context=True
                ):
                    return True
                if is_typing_union(get_origin(_resolver.unwrap_annotated(o))):
                    if recursive_contains_struct_type(get_args(o)):  # type: ignore
                        return True
            return False

        if not found_subcommand_conf and not recursive_contains_struct_type(options):
            return None

        # Get subcommand configurations from `tyro.conf.subcommand()`.
        subcommand_config_from_name: Dict[str, _confstruct._SubcommandConfig] = {}
        subcommand_type_from_name: Dict[str, type] = {}
        subcommand_names: list[str] = []
        for option in options:
            option_unwrapped, found_subcommand_configs = _resolver.unwrap_annotated(
                option, _confstruct._SubcommandConfig
            )
            subcommand_name = _strings.subparser_name_from_type(
                (
                    ""
                    if _markers.OmitSubcommandPrefixes in field.markers
                    else extern_prefix
                ),
                cast(type, option),
            )
            subcommand_names.append(subcommand_name)
            if subcommand_name in subcommand_type_from_name:
                # Raise a warning that the subcommand already exists
                original_type = subcommand_type_from_name[subcommand_name]
                original_type_full_name = (
                    f"{original_type.__module__}.{original_type.__name__}"
                )
                new_type_full_name = (
                    f"{option_unwrapped.__module__}.{option_unwrapped.__name__}"
                    if option_unwrapped is not None
                    else "none"
                )

                warnings.warn(
                    f"Duplicate subcommand name detected: '{subcommand_name}' is already used for "
                    f"{original_type_full_name} but will be overwritten by {new_type_full_name}. "
                    f"Only the last type ({new_type_full_name}) will be accessible via this subcommand. "
                    f"Consider using distinct class names or use tyro.conf.subcommand() to specify "
                    f"explicit subcommand names."
                )
            if len(found_subcommand_configs) != 0:
                # Explicitly annotated default.
                assert len(found_subcommand_configs) == 1, (
                    f"Expected only one subcommand config, but {subcommand_name} has"
                    f" {len(found_subcommand_configs)}."
                )
                subcommand_config_from_name[subcommand_name] = found_subcommand_configs[
                    0
                ]
            subcommand_type_from_name[subcommand_name] = cast(type, option)

        # If a field default is provided, try to find a matching subcommand name.
        # Note: EXCLUDE_FROM_CALL (from TypedDict total=False or NotRequired[]) is
        # a sentinel that means no default was provided, so we skip matching.
        default_name = (
            _subcommand_matching.match_subcommand(
                field.default,
                subcommand_config_from_name,
                subcommand_type_from_name,
                extern_prefix,
            )
            if not _singleton.is_sentinel(field.default)
            else None
        )

        if (
            default_name is not None
            and _markers.NewSubcommandForDefaults in field.markers
        ):
            # Create a new "default" subcommand instead of matching to an
            # existing one. This preserves original subcommand defaults.
            default_subcommand_name = (
                "default"
                if _markers.OmitSubcommandPrefixes in field.markers
                else f"{extern_prefix}:default"
            )
            assert default_subcommand_name not in subcommand_type_from_name, (
                f"Cannot create new default subcommand '{default_subcommand_name}' "
                f"because it already exists."
            )

            # Add to tracking structures. We inherit type from matched default
            # subcommand.
            default_config = _confstruct._SubcommandConfig(
                name="default",
                description="",
                default=field.default,
                prefix_name=True,
                constructor_factory=None,
            )
            default_type = subcommand_type_from_name[default_name]
            subcommand_config_from_name[default_subcommand_name] = default_config
            subcommand_type_from_name[default_subcommand_name] = default_type

            # Add matching pair of subcommand name and type.
            subcommand_names.append(default_subcommand_name)
            options.append(default_type)

            default_name = default_subcommand_name

        # Handle `tyro.conf.AvoidSubcommands` with a default value.
        if default_name is not None and _markers.AvoidSubcommands in field.markers:
            return ParserSpecification.from_callable_or_type(
                subcommand_type_from_name[default_name],
                markers=field.markers,
                description=None,
                parent_classes=parent_classes,
                default_instance=field.default,
                intern_prefix=intern_prefix,
                extern_prefix=extern_prefix,
                subcommand_prefix=extern_prefix,
                support_single_arg_types=True,
                prog_suffix=prog_suffix,
            )

        # Add subcommands for each option.
        parser_from_name: Dict[str, LazyParserSpecification] = {}
        for option, subcommand_name in zip(options, subcommand_names):
            # Get a subcommand config: either pulled from the type annotations or the
            # field default.
            if subcommand_name in subcommand_config_from_name:
                subcommand_config = subcommand_config_from_name[subcommand_name]
            else:
                subcommand_config = _confstruct._SubcommandConfig(
                    "unused",
                    description=None,
                    default=_singleton.MISSING_NONPROP,
                    prefix_name=True,
                    constructor_factory=None,
                )

            # If names match, borrow subcommand default from field default.
            if default_name == subcommand_name and not _singleton.is_missing(
                field.default
            ):
                subcommand_config = dataclasses.replace(
                    subcommand_config, default=field.default
                )

            # Strip the subcommand config from the option type.
            # Relevant: https://github.com/brentyi/tyro/pull/117
            option_unwrapped, annotations = _resolver.unwrap_annotated(option, "all")
            annotations = tuple(
                a
                for a in annotations
                if not isinstance(a, _confstruct._SubcommandConfig)
            )
            if _markers.Suppress in annotations:
                continue

            # Skip None options when DisallowNone is present.
            # This follows the same pattern as regular field DisallowNone handling.
            if (
                option_unwrapped is type(None)
                and _markers.DisallowNone in field.markers
            ):
                continue

            if len(annotations) == 0:
                option = option_unwrapped
            else:
                option = Annotated[(option_unwrapped,) + annotations]  # type: ignore

            # Extract description early for fast help text generation.
            # If no explicit description, get it from the callable's docstring.
            description_for_help = subcommand_config.description
            if option_unwrapped is type(None):
                description_for_help = ""
            elif description_for_help is None:
                description_for_help = _docstrings.get_callable_description(
                    option_unwrapped
                )

            # Create lazy parser: defer expensive parsing until actually needed.
            def parser_factory(
                option_captured: Any = option,
                markers_captured: Set[_markers._Marker] = field.markers,
                subcommand_config_captured: _confstruct._SubcommandConfig = subcommand_config,
                parent_classes_captured: Set[Type[Any]] = parent_classes,
                intern_prefix_captured: str = intern_prefix,
                extern_prefix_captured: str = extern_prefix,
                prog_suffix_captured: str = prog_suffix,
                subcommand_name_captured: str = subcommand_name,
                field_markers_captured: Set[_markers._Marker] = field.markers,
            ) -> ParserSpecification:
                with _fields.FieldDefinition.marker_context(
                    tuple(field_markers_captured)
                ):
                    subparser = ParserSpecification.from_callable_or_type(
                        option_captured,  # type: ignore
                        markers=markers_captured,
                        description=subcommand_config_captured.description,
                        parent_classes=parent_classes_captured,
                        default_instance=subcommand_config_captured.default,
                        intern_prefix=intern_prefix_captured,
                        extern_prefix=extern_prefix_captured,
                        subcommand_prefix=extern_prefix_captured,
                        support_single_arg_types=True,
                        prog_suffix=subcommand_name_captured
                        if prog_suffix_captured == ""
                        else prog_suffix_captured + " " + subcommand_name_captured,
                    )
                # Apply prefix to helptext in nested classes in subparsers.
                subparser = dataclasses.replace(
                    subparser,
                    helptext_from_intern_prefixed_field_name={
                        _strings.make_field_name([intern_prefix_captured, k]): v
                        for k, v in subparser.helptext_from_intern_prefixed_field_name.items()
                    },
                )
                return subparser

            parser_from_name[subcommand_name] = LazyParserSpecification(
                description=_strings.remove_single_line_breaks(description_for_help),
                _factory=parser_factory,  # type: ignore
            )

        # Default parser was suppressed!
        if default_name not in parser_from_name:
            default_name = None

        # Required if a default is passed in, but the default value has missing
        # parameters.
        default_parser = None
        if default_name is None:
            # Special case for DisallowNone: if the default would be None but we
            # suppressed the None subcommand due to DisallowNone, keep it optional.
            if field.default is None and _markers.DisallowNone in field.markers:
                required = False
            else:
                # If the default is EXCLUDE_FROM_CALL (from TypedDict total=False or
                # NotRequired[Union[...]]), the subparser is optional. When no subcommand
                # is selected, the field will be excluded from the result (see _calling.py).
                required = field.default is not _singleton.EXCLUDE_FROM_CALL
        else:
            required = False
            # Evaluate the lazy parser to check for required args/subparsers.
            default_parser_evaluated = parser_from_name[default_name].evaluate()
            # Error should have been caught earlier.
            assert not isinstance(
                default_parser_evaluated, UnsupportedTypeAnnotationError
            ), "Unexpected UnsupportedTypeAnnotationError in backend"

            # If there are any required arguments.
            if any(
                map(lambda arg: arg.lowered.required, default_parser_evaluated.args)
            ):
                required = True
                default_parser = None
            # If there are any required subparsers.
            elif any(
                subparser_spec.required
                for subparser_spec in default_parser_evaluated.subparsers_from_intern_prefix.values()
            ):
                required = True
                default_parser = None
            else:
                default_parser = default_parser_evaluated

        return SubparsersSpecification(
            # If we wanted, we could add information about the default instance
            # automatically, as is done for normal fields. But for now we just rely on
            # the user to include it in the docstring.
            # Keep description lazy - don't evaluate yet.
            description=field.helptext,
            parser_from_name=parser_from_name,
            default_name=default_name,
            default_parser=default_parser,
            intern_prefix=intern_prefix,
            extern_prefix=extern_prefix,
            required=required,
            default_instance=field.default,
            options=tuple(options),
            prog_suffix=prog_suffix,
        )



"""Tests for tuples containing unions with different lengths."""

from __future__ import annotations

import dataclasses
from typing import List, Literal, Optional, Tuple

import pytest

import tyro
from tyro.constructors._primitive_spec import (
    PrimitiveConstructorSpec,
    PrimitiveTypeInfo,
)


@dataclasses.dataclass
class _TestInner:
    values: Tuple[float, float] | Tuple[float, float, float]


@dataclasses.dataclass
class _TestOuter:
    inner: _TestInner
    extra: Tuple[str] | Tuple[str, str]


def test_basic_union_of_tuples():
    """Test basic union of tuples with different lengths."""

    def main(x: Tuple[int, int] | Tuple[int, int, int]) -> Tuple[int, ...]:
        return x

    # Test with 2 elements.
    assert tyro.cli(main, args=["--x", "1", "2"]) == (1, 2)

    # Test with 3 elements.
    assert tyro.cli(main, args=["--x", "1", "2", "3"]) == (1, 2, 3)

    # Test with wrong number of elements.
    with pytest.raises(SystemExit):
        tyro.cli(main, args=["--x", "1"])

    with pytest.raises(SystemExit):
        tyro.cli(main, args=["--x", "1", "2", "3", "4"])


def test_union_of_tuples_with_strings():
    """Test union of tuples with different types."""

    def main(x: Tuple[str, str] | Tuple[int, int, int]) -> tuple:
        return x

    # Test with 2 strings.
    assert tyro.cli(main, args=["--x", "hello", "world"]) == ("hello", "world")

    # Test with 3 integers.
    assert tyro.cli(main, args=["--x", "1", "2", "3"]) == (1, 2, 3)


def test_nested_tuple_with_union() -> None:
    """Test nested tuple containing a union with different lengths."""

    @dataclasses.dataclass
    class Config:
        x: Tuple[Tuple[int, int] | Tuple[int, int, int]]

    # Test with inner tuple of 2 elements.
    result = tyro.cli(Config, args=["--x", "1", "2"])
    assert result.x == ((1, 2),)

    # Test with inner tuple of 3 elements.
    result = tyro.cli(Config, args=["--x", "1", "2", "3"])
    assert result.x == ((1, 2, 3),)


def test_optional_nested_tuple_with_union() -> None:
    """Test optional nested tuple containing a union."""

    @dataclasses.dataclass
    class Config:
        x: Optional[Tuple[Tuple[int, int] | Tuple[int, int, int]]]

    # Test with None.
    result = tyro.cli(Config, args=["--x", "None"])
    assert result.x is None

    # Test with tuple of 2 elements.
    result = tyro.cli(Config, args=["--x", "1", "2"])
    assert result.x == ((1, 2),)

    # Test with tuple of 3 elements.
    result = tyro.cli(Config, args=["--x", "1", "2", "3"])
    assert result.x == ((1, 2, 3),)


def test_multiple_unions_in_tuple():
    """Test tuple with multiple union elements."""

    def main(x: Tuple[Tuple[int] | Tuple[int, int], str]) -> tuple:
        return x

    # Test with single int and string.
    assert tyro.cli(main, args=["--x", "1", "hello"]) == ((1,), "hello")

    # Test with pair of ints and string.
    assert tyro.cli(main, args=["--x", "1", "2", "world"]) == ((1, 2), "world")


def test_deeply_nested_unions():
    """Test deeply nested structures with variable-length unions."""

    # Test a simple list of variable-length tuples.
    def main(x: List[Tuple[int, int] | Tuple[int, int, int]]) -> list:
        return x

    # Test with a single tuple.
    result = tyro.cli(main, args=["--x", "1", "2"])
    assert result == [(1, 2)]

    # Test with a different length tuple.
    result = tyro.cli(main, args=["--x", "3", "4", "5"])
    assert result == [(3, 4, 5)]


def test_union_with_literals():
    """Test union of tuples containing literals."""

    def main(
        x: tuple[Literal["a"], int] | tuple[Literal["b"], int, int],
    ) -> tuple:
        return x

    # Test with literal "a".
    assert tyro.cli(main, args=["--x", "a", "1"]) == ("a", 1)

    # Test with literal "b".
    assert tyro.cli(main, args=["--x", "b", "1", "2"]) == ("b", 1, 2)

    # Test with wrong literal.
    with pytest.raises(SystemExit):
        tyro.cli(main, args=["--x", "c", "1"])


def test_complex_nested_structure():
    """Test complex nested structure with variable-length tuples."""

    # Test with all minimum lengths.
    result = tyro.cli(
        _TestOuter, args=["--inner.values", "1.0", "2.0", "--extra", "hello"]
    )
    assert result.inner.values == (1.0, 2.0)
    assert result.extra == ("hello",)

    # Test with all maximum lengths.
    result = tyro.cli(
        _TestOuter,
        args=["--inner.values", "1.0", "2.0", "3.0", "--extra", "hello", "world"],
    )
    assert result.inner.values == (1.0, 2.0, 3.0)
    assert result.extra == ("hello", "world")


def test_helptext_for_variable_length_tuples():
    """Test that help text is generated correctly."""

    def main(x: Tuple[int, int] | Tuple[int, int, int]) -> tuple:
        """Test function.

        Args:
            x: A tuple with either 2 or 3 integers.
        """
        return x

    # Just check that help doesn't crash.
    with pytest.raises(SystemExit) as exc_info:
        tyro.cli(main, args=["--help"])
    assert exc_info.value.code == 0


def test_primitive_spec_nargs_computation():
    """Test that nargs is computed correctly for variable-length tuples."""
    from tyro.constructors._registry import ConstructorRegistry

    registry = ConstructorRegistry()

    # Test simple union of tuples.
    type_info = PrimitiveTypeInfo.make(
        Tuple[int, int] | Tuple[int, int, int],  # type: ignore
        parent_markers=set(),
    )
    spec = registry.get_primitive_spec(type_info)
    assert isinstance(spec, PrimitiveConstructorSpec)
    assert spec.nargs == (2, 3)

    # Test nested tuple with union.
    type_info = PrimitiveTypeInfo.make(
        Tuple[Tuple[int, int] | Tuple[int, int, int]], parent_markers=set()
    )
    spec = registry.get_primitive_spec(type_info)
    assert isinstance(spec, PrimitiveConstructorSpec)
    assert spec.nargs == (2, 3)


def test_error_on_star_nargs():
    """Test that star nargs in tuples raises appropriate error."""
    from tyro.constructors._primitive_spec import UnsupportedTypeAnnotationError
    from tyro.constructors._registry import ConstructorRegistry

    registry = ConstructorRegistry()

    # This should raise an error because list[int] has nargs="*".
    type_info = PrimitiveTypeInfo.make(Tuple[List[int], int], parent_markers=set())
    spec = registry.get_primitive_spec(type_info)
    assert not isinstance(spec, UnsupportedTypeAnnotationError)
    assert spec.nargs == "*"


def test_backtracking_with_choices():
    """Test that backtracking works correctly with choices."""

    def main(
        x: Tuple[Literal["a", "b"], int] | Tuple[Literal["c"], int, int],
    ) -> tuple:
        return x

    # Test valid choices.
    assert tyro.cli(main, args=["--x", "a", "1"]) == ("a", 1)
    assert tyro.cli(main, args=["--x", "b", "1"]) == ("b", 1)
    assert tyro.cli(main, args=["--x", "c", "1", "2"]) == ("c", 1, 2)

    # Test invalid choice.
    with pytest.raises(SystemExit):
        tyro.cli(main, args=["--x", "d", "1"])


def test_ambiguous_parsing():
    """Test cases where parsing could be ambiguous."""

    def main(
        x: Tuple[Tuple[int] | Tuple[int, int], Tuple[int] | Tuple[int, int]],
    ) -> tuple:
        return x

    # This should parse as ((1,), (2, 3)).
    assert tyro.cli(main, args=["--x", "1", "2", "3"]) == ((1,), (2, 3))

    # This should parse as ((1, 2), (3,)).
    assert tyro.cli(main, args=["--x", "1", "2", "3"]) == ((1,), (2, 3))
    # Note: Due to greedy left-to-right parsing, this will actually be ((1,), (2, 3)).
    # This is expected behavior with the backtracking algorithm.


def test_multiple_variable_length_elements():
    """Test tuple with multiple variable-length elements."""

    def main(
        x: Tuple[
            Tuple[int, int] | Tuple[int, int, int],
            Tuple[str] | Tuple[str, str],
            Tuple[float, float, float] | Tuple[float, float, float, float],
        ],
    ) -> tuple:
        return x

    # Test with all minimum lengths.
    result = tyro.cli(main, args=["--x", "1", "2", "hello", "1.0", "2.0", "3.0"])
    assert result == ((1, 2), ("hello",), (1.0, 2.0, 3.0))

    # Test with mixed lengths.
    result = tyro.cli(
        main, args=["--x", "1", "2", "3", "hello", "world", "1.0", "2.0", "3.0", "4.0"]
    )
    assert result == ((1, 2, 3), ("hello", "world"), (1.0, 2.0, 3.0, 4.0))


if __name__ == "__main__":
    # Run a simple test when executed directly.
    test_basic_union_of_tuples()
    print("Basic test passed!")



import dataclasses
import time
from typing import Annotated, Literal

import tyro


@dataclasses.dataclass(frozen=True)
class AlgorithmConfig:
    flow_steps: int = 10
    output_mode: Literal["u", "u_but_supervise_as_eps"] = "u_but_supervise_as_eps"
    timestep_embed_dim: int = 8
    n_samples_per_action: int = 8
    average_losses_before_exp: bool = True
    discretize_t_for_training: bool = True
    feather_std: float = 0.0
    policy_mlp_output_scale: float = 0.25
    trust_region_mode: Literal["ppo", "spo"] = "ppo"
    loss_mode: Literal["fpo", "denoising_mdp"] = "fpo"
    final_steps_only: bool = False
    sde_sigma: float = 0.0
    clipping_epsilon: float = 0.05
    batch_size: int = 1024
    discounting: float = 0.995
    episode_length: int = 1000
    learning_rate: float = 3e-4
    normalize_observations: bool = True
    num_envs: int = 2048
    num_evals: int = 10
    num_minibatches: int = 32
    num_timesteps: int = 60_000_000
    num_updates_per_batch: int = 16
    reward_scaling: float = 10.0
    unroll_length: int = 30
    gae_lambda: float = 0.95
    normalize_advantage: bool = True
    value_loss_coeff: float = 0.25


@dataclasses.dataclass(frozen=True)
class EnvironmentConfig:
    env_name: str = "HalfCheetah-v4"
    env_seed: int = 0
    env_num_threads: int = 1
    env_max_episode_steps: int = 1000
    env_action_repeat: int = 1
    env_frame_stack: int = 1
    env_observation_normalization: bool = True
    env_reward_normalization: bool = True


@dataclasses.dataclass(frozen=True)
class RewardConfig:
    reward_scale: float = 1.0
    reward_shift: float = 0.0
    reward_normalization: bool = True
    reward_clip: float = 10.0


@dataclasses.dataclass(frozen=True)
class LoggingConfig:
    log_dir: str = "logs"
    log_interval: int = 1000
    save_model_interval: int = 10000
    eval_interval: int = 5000
    eval_episodes: int = 10
    tensorboard_log: bool = True


def main(n: int = 4) -> None:
    @dataclasses.dataclass(frozen=True)
    class ExperimentConfig:
        algorithm: Annotated[
            AlgorithmConfig,
            tyro.conf.arg(
                constructor=tyro.extras.subcommand_type_from_defaults(
                    {str(i): AlgorithmConfig(flow_steps=i) for i in range(n)}
                )
            ),
        ] = AlgorithmConfig()
        env: Annotated[
            EnvironmentConfig,
            tyro.conf.arg(
                constructor=tyro.extras.subcommand_type_from_defaults(
                    {str(i): EnvironmentConfig(env_name=f"Env-{i}") for i in range(n)}
                )
            ),
        ] = EnvironmentConfig()
        reward: Annotated[
            RewardConfig,
            tyro.conf.arg(
                constructor=tyro.extras.subcommand_type_from_defaults(
                    {str(i): RewardConfig(reward_scale=i) for i in range(n)}
                )
            ),
        ] = RewardConfig()
        logging: Annotated[
            LoggingConfig,
            tyro.conf.arg(
                constructor=tyro.extras.subcommand_type_from_defaults(
                    {str(i): LoggingConfig(log_dir=f"logs_{i}") for i in range(n)}
                )
            ),
        ] = LoggingConfig()
        logging2: Annotated[
            LoggingConfig,
            tyro.conf.arg(
                constructor=tyro.extras.subcommand_type_from_defaults(
                    {str(i): LoggingConfig(log_dir=f"logs_{i}") for i in range(n)}
                )
            ),
        ] = LoggingConfig()
        logging3: Annotated[
            LoggingConfig,
            tyro.conf.arg(
                constructor=tyro.extras.subcommand_type_from_defaults(
                    {str(i): LoggingConfig(log_dir=f"logs_{i}") for i in range(n)}
                )
            ),
        ] = LoggingConfig()

    start = time.perf_counter()
    tyro.cli(ExperimentConfig, args=[])
    print(f"Total time taken: {(time.perf_counter() - start) * 1000:.1f}ms")


if __name__ == "__main__":
    tyro.cli(main)



"""Shell completion generation for tyro."""

from __future__ import annotations

from ._base import CompletionGenerator
from ._tyro_bash import TyroBashCompletionGenerator
from ._tyro_zsh import TyroZshCompletionGenerator

__all__ = [
    "CompletionGenerator",
    "TyroBashCompletionGenerator",
    "TyroZshCompletionGenerator",
]



import sys
from typing import Any

if sys.version_info >= (3, 11):
    SubclassableAny = Any
else:  # pragma: no cover
    from typing_extensions import Any as SubclassableAny


class Singleton:
    # Singleton pattern.
    # https://www.python.org/download/releases/2.2/descrintro/#__new__
    def __new__(cls, *args, **kwds):
        it = cls.__dict__.get("__it__")
        if it is not None:
            return it
        cls.__it__ = it = object.__new__(cls)
        it.init(*args, **kwds)
        return it

    def init(self, *args, **kwds):
        pass


# We subclass `Any` to prevent typeguard from failing for MISSING types.
#
# https://github.com/agronholm/typeguard/blob/dd98a9a0ff050166716120cc8614fa90d710a879/src/typeguard/_checkers.py#L933-L935


class PropagatingMissingType(Singleton, SubclassableAny):
    """Type for the :data:`tyro.MISSING` singleton."""

    def __repr__(self) -> str:
        return "tyro.MISSING"


class NonpropagatingMissingType(Singleton, SubclassableAny):
    """Type for the :data:`tyro.MISSING_NONPROP` singleton."""

    def __repr__(self) -> str:
        return "tyro.MISSING_NONPROP"


class ExcludeFromCallType(Singleton):
    pass


class NotRequiredButWeDontKnowTheValueType(Singleton):
    pass


# We have two types of missing sentinels: a propagating missing value, which when set as
# a default will set all child values of nested structures as missing as well, and a
# nonpropagating missing sentinel, which does not override child defaults.
MISSING: Any = PropagatingMissingType()
"""Sentinel value to mark default values as missing. Can be used to mark fields
passed in via `default=` for `tyro.cli()` as required.

When used, the 'missing' semantics propagate to children. For example, if we write:

.. code-block:: python

    def main(inner: Dataclass = tyro.MISSING) -> None:
        ...

    tyro.cli(main)

then all fields belonging to ``Dataclass`` will be marked as missing, even if a
default exists in the dataclass definition.
"""
MISSING_NONPROP: Any = NonpropagatingMissingType()
"""Non-propagating version of :data:`tyro.MISSING`.

When used, the 'missing' semantics do not propagate to children. For example:

.. code-block:: python

    def main(inner: Dataclass = tyro.MISSING_NONPROP) -> None:
        ...

    tyro.cli(main)

is equivalent to:

.. code-block:: python

    def main(inner: Dataclass) -> None:
        ...

    tyro.cli(main)

where default values for fields belonging to ``Dataclass`` will be taken from
the dataclass definition.
"""


EXCLUDE_FROM_CALL = ExcludeFromCallType()
"""Singleton indicating that an argument should not be passed into a field
constructor. This is used for :py:class:`typing.TypedDict`."""


def is_missing(value: Any) -> bool:
    """Check if a value is a missing sentinel (MISSING or MISSING_NONPROP).

    Uses identity checks to avoid issues with types like numpy arrays that have
    ambiguous truth values when compared with `==` or `in`.
    """
    return value is MISSING or value is MISSING_NONPROP


def is_sentinel(value: Any) -> bool:
    """Check if a value is any default sentinel (MISSING, MISSING_NONPROP, or EXCLUDE_FROM_CALL).

    Uses identity checks to avoid issues with types like numpy arrays that have
    ambiguous truth values when compared with `==` or `in`.
    """
    return value is MISSING or value is MISSING_NONPROP or value is EXCLUDE_FROM_CALL



"""Adapted from @mirceamironenco.

https://github.com/brentyi/tyro/issues/340"""

from __future__ import annotations

from dataclasses import dataclass

from helptext_utils import get_helptext_with_checks

import tyro


@dataclass(frozen=True)
class HFSFTDatasetConfig:
    name: str
    split: str | None = None
    data_files: str | None = None

    completions_only: bool = True
    packed_seqlen: int | None = None
    apply_chat_template: bool = False
    max_seq_len: int | None = None

    columns: tuple[str, str, str | None, str | None] | None = None
    """instruction, completion, input, system"""


@dataclass
class Config:
    train_data: HFSFTDatasetConfig = HFSFTDatasetConfig(
        name="yahma/alpaca-cleaned",
        completions_only=True,
        packed_seqlen=4097,
        max_seq_len=2048,
        apply_chat_template=True,
        columns=("instruction", "output", "input", None),
    )


def test_mixed_primitive() -> None:
    assert (
        tyro.cli(
            Config,
            args=["--train-data.name", "foo", "--train-data.completions-only"],
        ).train_data.name
        == "foo"
    )
    assert "None}|{STR STR {None}|STR {None}|STR}" in get_helptext_with_checks(Config)



# mypy: ignore-errors
"""Tests for add_help parameter functionality."""

import dataclasses
import io
import sys

import pytest

import tyro


@dataclasses.dataclass
class SimpleConfig:
    """A simple configuration class."""

    value: int
    name: str = "default"


def simple_function(x: int, y: str = "hello") -> str:
    """A simple function for testing."""
    return f"{x}:{y}"


def test_cli_add_help_false() -> None:
    """Test that add_help=False prevents help from being added."""
    # Should raise an error when --help is provided with add_help=False
    with pytest.raises(SystemExit) as excinfo:
        tyro.cli(SimpleConfig, args=["--help"], add_help=False)
    assert excinfo.value.code == 2  # Should fail with parsing error, not help display


def test_cli_add_help_true() -> None:
    """Test that add_help=True (default) allows help."""
    # Should show help and exit with code 0
    with pytest.raises(SystemExit) as excinfo:
        # Redirect stdout to capture help output
        old_stdout = sys.stdout
        sys.stdout = io.StringIO()
        try:
            tyro.cli(SimpleConfig, args=["--help"], add_help=True)
        finally:
            sys.stdout = old_stdout
    assert excinfo.value.code == 0  # Should exit cleanly after showing help


def test_cli_default_add_help() -> None:
    """Test that add_help defaults to True."""
    # Should show help and exit with code 0
    with pytest.raises(SystemExit) as excinfo:
        old_stdout = sys.stdout
        sys.stdout = io.StringIO()
        try:
            tyro.cli(SimpleConfig, args=["--help"])
        finally:
            sys.stdout = old_stdout
    assert excinfo.value.code == 0


def test_get_parser_add_help_false() -> None:
    """Test that get_parser with add_help=False doesn't add help option."""
    parser = tyro.extras.get_parser(SimpleConfig, add_help=False)
    assert "-h" not in parser._option_string_actions
    assert "--help" not in parser._option_string_actions


def test_get_parser_add_help_true() -> None:
    """Test that get_parser with add_help=True adds help option."""
    parser = tyro.extras.get_parser(SimpleConfig, add_help=True)
    assert (
        "-h" in parser._option_string_actions
        or "--help" in parser._option_string_actions
    )


def test_get_parser_default_add_help() -> None:
    """Test that get_parser defaults to add_help=True."""
    parser = tyro.extras.get_parser(SimpleConfig)
    assert (
        "-h" in parser._option_string_actions
        or "--help" in parser._option_string_actions
    )


def test_function_cli_add_help() -> None:
    """Test add_help works with function targets."""
    # Test with add_help=False
    result = tyro.cli(simple_function, args=["--x", "42"], add_help=False)
    assert result == "42:hello"

    # Test that --help fails with add_help=False
    with pytest.raises(SystemExit) as excinfo:
        tyro.cli(simple_function, args=["--help"], add_help=False)
    assert excinfo.value.code == 2


def test_subcommand_app_add_help() -> None:
    """Test add_help parameter with SubcommandApp."""
    from tyro.extras import SubcommandApp

    app = SubcommandApp()

    @app.command
    def cmd1(x: int) -> int:
        return x

    @app.command
    def cmd2(y: str) -> str:
        return y

    # Test with add_help=False
    result = app.cli(args=["cmd1", "--x", "5"], add_help=False)
    assert result == 5

    # Test that --help fails with add_help=False
    with pytest.raises(SystemExit) as excinfo:
        app.cli(args=["--help"], add_help=False)
    assert excinfo.value.code == 2


def test_subcommand_cli_from_dict_add_help() -> None:
    """Test add_help parameter with subcommand_cli_from_dict."""

    def cmd1(x: int) -> int:
        return x * 2

    def cmd2(y: str) -> str:
        return y.upper()

    subcommands = {"multiply": cmd1, "uppercase": cmd2}

    # Test with add_help=False
    result = tyro.extras.subcommand_cli_from_dict(
        subcommands, args=["multiply", "--x", "3"], add_help=False
    )
    assert result == 6

    # Test that --help fails with add_help=False
    with pytest.raises(SystemExit) as excinfo:
        tyro.extras.subcommand_cli_from_dict(
            subcommands, args=["--help"], add_help=False
        )
    assert excinfo.value.code == 2


def test_overridable_config_cli_add_help() -> None:
    """Test add_help parameter with overridable_config_cli."""

    @dataclasses.dataclass
    class Config:
        a: int
        b: str

    configs = {
        "small": ("Small config", Config(1, "small")),
        "big": ("Big config", Config(100, "big")),
    }

    # Test with add_help=False
    result = tyro.extras.overridable_config_cli(configs, args=["small"], add_help=False)
    assert result.a == 1
    assert result.b == "small"

    # Test that --help fails with add_help=False
    with pytest.raises(SystemExit) as excinfo:
        tyro.extras.overridable_config_cli(configs, args=["--help"], add_help=False)
    assert excinfo.value.code == 2


def test_subparsers_respect_add_help() -> None:
    """Test that subparsers inherit the add_help setting from parent parser."""
    from typing import Union

    @dataclasses.dataclass
    class ConfigA:
        value_a: int = 1

    @dataclasses.dataclass
    class ConfigB:
        value_b: str = "test"

    # Test with add_help=False - subparsers should also not have help
    with pytest.raises(SystemExit) as excinfo:
        tyro.cli(Union[ConfigA, ConfigB], args=["config-a", "--help"], add_help=False)
    assert excinfo.value.code == 2  # Should fail, not show help

    # Test with add_help=True - subparsers should have help
    with pytest.raises(SystemExit) as excinfo:
        old_stdout = sys.stdout
        sys.stdout = io.StringIO()
        try:
            tyro.cli(
                Union[ConfigA, ConfigB], args=["config-a", "--help"], add_help=True
            )
        finally:
            sys.stdout = old_stdout
    assert excinfo.value.code == 0  # Should show help

    # Test that subparser works normally with add_help=False
    result = tyro.cli(Union[ConfigA, ConfigB], args=["config-a"], add_help=False)
    assert isinstance(result, ConfigA)
    assert result.value_a == 1


def test_return_unknown_args_with_add_help_false() -> None:
    """Test that --help/-h are returned as unknown args when add_help=False and return_unknown_args=True."""

    @dataclasses.dataclass
    class Config:
        value: int = 5

    # Test that --help is returned in unknown args
    result, unknown = tyro.cli(
        Config, args=["--help"], add_help=False, return_unknown_args=True
    )
    assert unknown == ["--help"]
    assert result.value == 5

    # Test that -h is returned in unknown args
    result, unknown = tyro.cli(
        Config, args=["-h"], add_help=False, return_unknown_args=True
    )
    assert unknown == ["-h"]
    assert result.value == 5

    # Test with both --help and other unknown args
    result, unknown = tyro.cli(
        Config,
        args=["--help", "--unknown", "arg"],
        add_help=False,
        return_unknown_args=True,
    )
    assert "--help" in unknown
    assert "--unknown" in unknown
    assert "arg" in unknown
    assert result.value == 5

    # Test that with add_help=True, --help still shows help (doesn't return unknown args)
    with pytest.raises(SystemExit) as excinfo:
        old_stdout = sys.stdout
        sys.stdout = io.StringIO()
        try:
            tyro.cli(Config, args=["--help"], add_help=True, return_unknown_args=True)
        finally:
            sys.stdout = old_stdout
    assert excinfo.value.code == 0  # Should exit cleanly after showing help


def test_error_messages_respect_add_help() -> None:
    """Test that error messages don't suggest --help when add_help=False."""
    import contextlib

    @dataclasses.dataclass
    class Config:
        required_field: int

    # Capture stderr to check error messages
    captured_output = io.StringIO()

    # Test with add_help=False - should not mention --help
    with pytest.raises(SystemExit):
        with contextlib.redirect_stderr(captured_output):
            tyro.cli(Config, args=[], add_help=False, console_outputs=True)

    error_message = captured_output.getvalue()
    assert "--help" not in error_message, (
        f"Error message should not mention --help when add_help=False. Got: {error_message}"
    )

    # Test with add_help=True - should mention --help
    captured_output = io.StringIO()
    with pytest.raises(SystemExit):
        with contextlib.redirect_stderr(captured_output):
            tyro.cli(Config, args=[], add_help=True, console_outputs=True)

    error_message = captured_output.getvalue()
    assert "--help" in error_message, (
        f"Error message should mention --help when add_help=True. Got: {error_message}"
    )



"""Utilities and constants for working with strings."""

import contextlib
import functools
import re
import textwrap
from typing import Iterable, List, Literal, Sequence, Tuple, Type

from typing_extensions import get_args, get_origin

from tyro._typing_compat import is_typing_union

from . import _resolver

DELIMETER: Literal["-", "_"] = "-"


@contextlib.contextmanager
def delimeter_context(delimeter: Literal["-", "_"]):
    """Context for setting the delimeter. Determines if `field_a` is populated as
    `--field-a` or `--field_a`. Not thread-safe."""
    global DELIMETER
    delimeter_restore = DELIMETER
    DELIMETER = delimeter
    yield
    DELIMETER = delimeter_restore


def get_delimeter() -> Literal["-", "_"]:
    """Get delimeter used to separate words."""
    return DELIMETER


def swap_delimeters(p: str) -> str:
    """Replace hyphens with underscores (or vice versa) except when at the start or end."""
    if get_delimeter() == "-":
        left_stripped = p.lstrip("_")
        right_stripped = left_stripped.rstrip("_")
        prefix = p[: len(p) - len(left_stripped)]
        suffix = left_stripped[len(right_stripped) :]
        p = prefix + right_stripped.replace("_", "-") + suffix
    else:
        left_stripped = p.lstrip("-")
        right_stripped = left_stripped.rstrip("-")
        prefix = p[: len(p) - len(left_stripped)]
        suffix = left_stripped[len(right_stripped) :]
        p = prefix + right_stripped.replace("-", "_") + suffix
    return p


def make_field_name(parts: Sequence[str]) -> str:
    """Join parts of a field name together. Used for nesting.

    ('parent_1', 'child') => 'parent-1.child'
    ('parents', '1', '_child_node') => 'parents.1._child-node'
    ('parents', '1', 'middle._child_node') => 'parents.1.middle._child-node'
    """
    out = ".".join(parts)
    return ".".join(swap_delimeters(part) for part in out.split(".") if len(part) > 0)


def make_subparser_dest(name: str) -> str:
    return f"{name} (positional)"


def dedent(text: str) -> str:
    """Same as textwrap.dedent, but ignores the first line."""
    first_line, line_break, rest = text.partition("\n")
    if line_break == "":
        return textwrap.dedent(text)
    return f"{first_line.strip()}\n{textwrap.dedent(rest)}"


def hyphen_separated_from_camel_case(name: str) -> str:
    out = (
        re.compile("((?<=[a-z0-9])[A-Z]|(?!^)[A-Z](?=[a-z]))")
        .sub(get_delimeter() + r"\1", name)
        .lower()
    )
    return out


def _subparser_name_from_type(cls: Type) -> Tuple[str, bool]:
    from .conf import _confstruct  # Prevent circular imports

    cls, type_from_typevar = _resolver.resolve_generic_types(cls)
    _, type_alias_breadcrumbs = _resolver.unwrap_annotated(
        cls, _resolver.TyroTypeAliasBreadCrumb
    )
    cls, found_subcommand_configs = _resolver.unwrap_annotated(
        cls, _confstruct._SubcommandConfig
    )

    # Subparser name from `tyro.conf.subcommand()`.
    found_name = None
    prefix_name = True
    if len(found_subcommand_configs) > 0:
        found_name = found_subcommand_configs[0].name
        prefix_name = found_subcommand_configs[0].prefix_name

    if found_name is not None:
        return found_name, prefix_name

    # Subparser name from type alias. This is lower priority thant he name from
    # `tyro.conf.subcommand()`.
    if len(type_alias_breadcrumbs) > 0:
        return (
            hyphen_separated_from_camel_case(type_alias_breadcrumbs[-1].name),
            prefix_name,
        )

    # Subparser name from class name.
    def get_name(cls: Type) -> str:
        orig = get_origin(cls)

        # Handle _SpecialForm version of Union, which doesn't have __name__,
        # normalize UnionType to Union, etc.
        orig_name: str | None
        if is_typing_union(orig):
            orig_name = "Union"
        else:
            orig_name = getattr(orig, "__name__", None)

        if orig_name is not None:
            parts = [orig_name]  # type: ignore
            parts.extend(map(get_name, get_args(cls)))
            parts = [hyphen_separated_from_camel_case(part) for part in parts]
            return get_delimeter().join(parts)
        elif hasattr(cls, "__name__"):
            return hyphen_separated_from_camel_case(cls.__name__)
        else:
            return hyphen_separated_from_camel_case(str(cls))

    if len(type_from_typevar) == 0:
        return get_name(cls), prefix_name  # type: ignore

    return (
        get_delimeter().join(
            [get_name(cls)]
            + list(
                map(
                    lambda x: _subparser_name_from_type(x)[0],
                    list(type_from_typevar.values()),
                )
            )
        ),
        prefix_name,
    )


def subparser_name_from_type(prefix: str, cls: Type) -> str:
    suffix, use_prefix = (
        _subparser_name_from_type(cls) if cls is not type(None) else ("none", True)
    )
    if len(prefix) == 0 or not use_prefix:
        return suffix

    if get_delimeter() == "-":
        return f"{prefix}:{make_field_name(suffix.split('.'))}"
    else:
        assert get_delimeter() == "_"
        return f"{prefix}:{suffix}"


@functools.lru_cache(maxsize=None)
def _get_ansi_pattern() -> re.Pattern:
    # https://stackoverflow.com/a/14693789
    return re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")


def strip_ansi_sequences(x: str):
    return _get_ansi_pattern().sub("", x)


def multi_metavar_from_single(single: str) -> str:
    if len(single) >= 32:
        # Shorten long metavars
        return f"[{single} [...]]"
    else:
        return f"[{single} [{single} ...]]"


def join_union_metavars(metavars: Iterable[str]) -> str:
    """Metavar generation helper for unions. Could be revisited.

    Examples:
        None, INT => NONE|INT
        {0,1,2}, {3,4} => {0,1,2,3,4}
        {0,1,2}, {3,4}, STR => {0,1,2,3,4}|STR
        {None}, INT [INT ...] => {None}|{INT [INT ...]}
        STR, INT [INT ...] => STR|{INT [INT ...]}
        STR, INT INT => STR|{INT INT}

    The curly brackets are unfortunately overloaded but alternatives all interfere with
    argparse internals.
    """
    metavars = tuple(metavars)
    merged_metavars = [metavars[0]]
    for i in range(1, len(metavars)):
        prev = merged_metavars[-1]
        curr = metavars[i]
        # Only merge curly brace expressions that represent literal choices (contain commas).
        # Don't merge if either contains spaces (indicating tuples/sequences) or nested braces.
        if (
            prev.startswith("{")
            and prev.endswith("}")
            and curr.startswith("{")
            and curr.endswith("}")
            and " " not in prev
            and " " not in curr
            and "{" not in prev[1:-1]
            and "{" not in curr[1:-1]
        ):
            merged_metavars[-1] = prev[:-1] + "," + curr[1:]
        else:
            merged_metavars.append(curr)

    for i, m in enumerate(merged_metavars):
        if " " in m:
            merged_metavars[i] = "{" + m + "}"

    return "|".join(merged_metavars)


def remove_single_line_breaks(helptext: str) -> str:
    lines = helptext.split("\n")
    output_parts: List[str] = []
    for line in lines:
        # Remove trailing whitespace.
        line = line.rstrip()

        # Empty line.
        if len(line) == 0:
            prev_is_break = len(output_parts) >= 1 and output_parts[-1] == "\n"
            if not prev_is_break:
                output_parts.append("\n")
            output_parts.append("\n")

        # Empty line.
        else:
            if not line[0].isalpha():
                output_parts.append("\n")
            prev_is_break = len(output_parts) >= 1 and output_parts[-1] == "\n"
            if len(output_parts) >= 1 and not prev_is_break:
                output_parts.append(" ")
            output_parts.append(line)

    return "".join(output_parts).rstrip()  # type: ignore



"""Tests to improve coverage of _tyro_backend.py."""

import dataclasses
from typing import List, Literal, Optional, Tuple, Union

import pytest
from typing_extensions import Annotated

import tyro


def test_positional():
    def main(
        x: int,
        y: int,
        /,
        # Note: it's generally a bad idea to have a mutable object (like a list) as a
        # default value. But it should still work.
        z: List[int] = [1, 2, 3],
    ) -> Tuple[int, int, int]:
        """main.

        Args:
            x: x
            y: y
            z: z

        Returns:
            Tuple[int, int, int]: Output.
        """
        return (x, y, z[0])

    assert tyro.cli(
        main, args="1 2 --z 3".split(" "), config=(tyro.conf.CascadeSubcommandArgs,)
    ) == (1, 2, 3)
    with pytest.raises(SystemExit):
        assert tyro.cli(
            main,
            args="--x 1 --y 2 --z 3".split(" "),
            config=(tyro.conf.CascadeSubcommandArgs,),
        ) == (1, 2, 3)


def test_cascade_misplaced_subcommand() -> None:
    """Test when subcommand appears after other arguments in cascaded mode."""

    @dataclasses.dataclass
    class SubcommandA:
        value: int = 1

    @dataclasses.dataclass
    class SubcommandB:
        value: int = 2

    @dataclasses.dataclass
    class Config:
        x: int
        subcommand: Union[SubcommandA, SubcommandB]

    if tyro._experimental_options["backend"] == "tyro":
        assert tyro.cli(
            tyro.conf.CascadeSubcommandArgs[Config],
            args=["--x", "5", "subcommand:subcommand-a"],
        ) == Config(x=5, subcommand=SubcommandA())
    else:
        with pytest.raises(SystemExit):
            tyro.cli(
                tyro.conf.CascadeSubcommandArgs[Config],
                args=["--x", "5", "subcommand:subcommand-a"],
            )


def test_cascade_optional_nargs() -> None:
    """Test nargs='?' handling in cascaded mode."""

    @dataclasses.dataclass
    class SubcommandA:
        required_value: int
        optional_value: Optional[int] = None

    @dataclasses.dataclass
    class SubcommandB:
        value: int = 1

    @dataclasses.dataclass
    class Config:
        subcommand: Union[SubcommandA, SubcommandB]

    # Test with the optional value provided.
    result = tyro.cli(
        tyro.conf.CascadeSubcommandArgs[Config],
        args=[
            "subcommand:subcommand-a",
            "--subcommand.required-value",
            "10",
            "--subcommand.optional-value",
            "42",
        ],
    )
    assert isinstance(result.subcommand, SubcommandA)
    assert result.subcommand.required_value == 10
    assert result.subcommand.optional_value == 42

    # Test without the optional value.
    result = tyro.cli(
        tyro.conf.CascadeSubcommandArgs[Config],
        args=["subcommand:subcommand-a", "--subcommand.required-value", "10"],
    )
    assert isinstance(result.subcommand, SubcommandA)
    assert result.subcommand.required_value == 10
    assert result.subcommand.optional_value is None


def test_cascade_positional_values() -> None:
    """Test variadic positional arguments in cascaded mode."""

    @dataclasses.dataclass
    class SubcommandA:
        values: Tuple[int, ...]

    @dataclasses.dataclass
    class SubcommandB:
        value: int = 1

    @dataclasses.dataclass
    class Config:
        subcommand: Union[SubcommandA, SubcommandB]

    result = tyro.cli(
        tyro.conf.CascadeSubcommandArgs[Config],
        args=[
            "subcommand:subcommand-a",
            "--subcommand.values",
            "1",
            "2",
            "3",
        ],
    )
    assert isinstance(result.subcommand, SubcommandA)
    assert result.subcommand.values == (1, 2, 3)


def test_cascade_nargs_plus_empty() -> None:
    """Test nargs='+' with values in cascaded mode."""

    @dataclasses.dataclass
    class SubcommandA:
        values: Tuple[int, ...]

    @dataclasses.dataclass
    class SubcommandB:
        value: int = 1

    @dataclasses.dataclass
    class Config:
        subcommand: Union[SubcommandA, SubcommandB]

    result = tyro.cli(
        tyro.conf.CascadeSubcommandArgs[Config],
        args=[
            "subcommand:subcommand-a",
            "--subcommand.values",
            "1",
            "2",
        ],
    )
    assert isinstance(result.subcommand, SubcommandA)
    assert result.subcommand.values == (1, 2)


def test_cascade_count_flag_basic() -> None:
    """Test count flags in cascaded mode."""

    @dataclasses.dataclass
    class Command:
        verbose: Annotated[
            tyro.conf.UseCounterAction[int], tyro.conf.arg(aliases=["-v"])
        ] = 0

    assert (
        tyro.cli(
            Command,
            config=(tyro.conf.UseCounterAction, tyro.conf.CascadeSubcommandArgs),
            args=["-vvv"],
        ).verbose
        == 3
    )
    assert (
        tyro.cli(
            Command,
            config=(tyro.conf.UseCounterAction, tyro.conf.CascadeSubcommandArgs),
            args=["-v", "-v", "-v"],
        ).verbose
        == 3
    )


def test_cascade_count_flag() -> None:
    """Test count flags in cascaded mode."""

    @dataclasses.dataclass
    class SubcommandA:
        verbose: tyro.conf.UseCounterAction[int] = 0

    @dataclasses.dataclass
    class SubcommandB:
        value: int = 1

    @dataclasses.dataclass
    class Config:
        subcommand: Union[SubcommandA, SubcommandB]

    result = tyro.cli(
        tyro.conf.CascadeSubcommandArgs[Config],
        args=[
            "subcommand:subcommand-a",
            "--subcommand.verbose",
            "--subcommand.verbose",
            "--subcommand.verbose",
        ],
    )
    assert isinstance(result.subcommand, SubcommandA)
    assert result.subcommand.verbose == 3


def test_cascade_flag_with_equals() -> None:
    """Test --flag=value syntax in cascaded mode."""

    @dataclasses.dataclass
    class SubcommandA:
        value: int = 1

    @dataclasses.dataclass
    class SubcommandB:
        value: int = 2

    @dataclasses.dataclass
    class Config:
        subcommand: Union[SubcommandA, SubcommandB]

    result = tyro.cli(
        tyro.conf.CascadeSubcommandArgs[Config],
        args=["subcommand:subcommand-a", "--subcommand.value=42"],
    )
    assert isinstance(result.subcommand, SubcommandA)
    assert result.subcommand.value == 42


def test_cascade_unknown_positional() -> None:
    """Test handling of unknown positional arguments in cascaded mode."""

    @dataclasses.dataclass
    class SubcommandA:
        value: int = 1

    @dataclasses.dataclass
    class SubcommandB:
        value: int = 2

    @dataclasses.dataclass
    class Config:
        subcommand: Union[SubcommandA, SubcommandB]

    # This should fail because there's an unknown positional argument.
    with pytest.raises(SystemExit):
        tyro.cli(
            tyro.conf.CascadeSubcommandArgs[Config],
            args=["subcommand:subcommand-a", "unknown_positional"],
        )


def test_cascade_default_subcommand() -> None:
    """Test default subcommand handling in cascaded mode."""

    @dataclasses.dataclass
    class SubcommandA:
        value: int = 1

    @dataclasses.dataclass
    class SubcommandB:
        value: int = 2

    @dataclasses.dataclass
    class Config:
        subcommand: Union[SubcommandA, SubcommandB] = dataclasses.field(
            default_factory=SubcommandA
        )

    result = tyro.cli(
        tyro.conf.CascadeSubcommandArgs[Config],
        args=[],
        default=Config(),
    )
    assert isinstance(result.subcommand, SubcommandA)
    assert result.subcommand.value == 1


def test_recursive_mode_default_subcommand() -> None:
    """Test default subcommand in recursive mode."""

    @dataclasses.dataclass
    class SubcommandA:
        value: int = 1

    @dataclasses.dataclass
    class SubcommandB:
        value: int = 2

    @dataclasses.dataclass
    class Config:
        subcommand: Union[SubcommandA, SubcommandB] = dataclasses.field(
            default_factory=SubcommandA
        )

    result = tyro.cli(
        Config,
        args=[],
        default=Config(),
    )
    assert isinstance(result.subcommand, SubcommandA)
    assert result.subcommand.value == 1


def test_recursive_mode_unknown_args() -> None:
    """Test unknown arguments in recursive mode cause an error."""

    @dataclasses.dataclass
    class Config:
        value: int = 1

    # Unknown args should cause an error.
    with pytest.raises(SystemExit):
        tyro.cli(
            Config,
            args=["--value", "42", "--unknown"],
        )


def test_choices_validation() -> None:
    """Test choices validation."""

    @dataclasses.dataclass
    class Config:
        option: Literal["a", "b", "c"]

    result = tyro.cli(Config, args=["--option", "a"])
    assert result.option == "a"

    # This should fail because "d" is not a valid choice.
    with pytest.raises(SystemExit):
        tyro.cli(Config, args=["--option", "d"])


def test_integer_nargs() -> None:
    """Test nargs with integer values."""

    @dataclasses.dataclass
    class Config:
        values: Tuple[int, int, int] = (1, 2, 3)

    result = tyro.cli(Config, args=["--values", "4", "5", "6"], default=Config())
    assert result.values == (4, 5, 6)

    # This should fail because we need exactly 3 values.
    with pytest.raises(SystemExit):
        tyro.cli(Config, args=["--values", "4", "5"], default=Config())


def test_help_flag_display() -> None:
    """Test help flag triggers help display."""

    @dataclasses.dataclass
    class Config:
        value: int = 1

    with pytest.raises(SystemExit) as exc_info:
        tyro.cli(Config, args=["--help"])

    assert exc_info.value.code == 0


def test_cascaded_help_display() -> None:
    """Test help display in cascaded mode."""

    @dataclasses.dataclass
    class SubcommandA:
        value: int = 1

    @dataclasses.dataclass
    class SubcommandB:
        value: int = 2

    @dataclasses.dataclass
    class Config:
        subcommand: Union[SubcommandA, SubcommandB]

    with pytest.raises(SystemExit) as exc_info:
        tyro.cli(
            tyro.conf.CascadeSubcommandArgs[Config],
            args=["subcommand:subcommand-a", "--help"],
        )

    assert exc_info.value.code == 0


def test_missing_required_args_error() -> None:
    """Test error when required arguments are missing."""

    @dataclasses.dataclass
    class Config:
        required_value: int

    with pytest.raises(SystemExit):
        tyro.cli(Config, args=[])


def test_boolean_flags() -> None:
    """Test boolean flags with store_true behavior."""

    @dataclasses.dataclass
    class Config:
        verbose: bool = False

    result = tyro.cli(Config, args=["--verbose"], default=Config())
    assert result.verbose is True

    result = tyro.cli(Config, args=[], default=Config())
    assert result.verbose is False


def test_nargs_star() -> None:
    """Test nargs='*' handling."""

    @dataclasses.dataclass
    class Config:
        values: List[int] = dataclasses.field(default_factory=list)

    result = tyro.cli(Config, args=["--values", "1", "2", "3"], default=Config())
    assert result.values == [1, 2, 3]

    result = tyro.cli(Config, args=[], default=Config())
    assert result.values == []


def test_register_parser_args() -> None:
    """Test registering arguments from parser specification."""

    @dataclasses.dataclass
    class Config:
        value1: int = 1
        value2: str = "test"
        value3: bool = False

    result = tyro.cli(
        Config,
        args=["--value1", "42", "--value2", "hello", "--value3"],
        default=Config(),
    )
    assert result.value1 == 42
    assert result.value2 == "hello"
    assert result.value3 is True


def test_add_help_flags() -> None:
    """Test help flags are registered."""

    @dataclasses.dataclass
    class Config:
        value: int = 1

    # Test both -h and --help.
    with pytest.raises(SystemExit) as exc_info:
        tyro.cli(Config, args=["-h"])
    assert exc_info.value.code == 0

    with pytest.raises(SystemExit) as exc_info:
        tyro.cli(Config, args=["--help"])
    assert exc_info.value.code == 0


def test_validate_required_args() -> None:
    """Test validation of required arguments."""

    @dataclasses.dataclass
    class Config:
        required1: int
        required2: str
        optional: bool = False

    # Should work with all required args.
    result = tyro.cli(
        Config,
        args=["--required1", "42", "--required2", "test"],
    )
    assert result.required1 == 42
    assert result.required2 == "test"
    assert result.optional is False

    # Should fail with missing required args.
    with pytest.raises(SystemExit):
        tyro.cli(Config, args=["--required1", "42"])


def test_nested_subcommands_cascaded() -> None:
    """Test nested subcommands in cascaded mode."""

    @dataclasses.dataclass
    class NestedA:
        value: int = 1

    @dataclasses.dataclass
    class NestedB:
        value: int = 2

    @dataclasses.dataclass
    class SubcommandA:
        x: int = 10
        nested: Union[NestedA, NestedB] = dataclasses.field(default_factory=NestedA)

    @dataclasses.dataclass
    class Config:
        subcommand: SubcommandA = dataclasses.field(default_factory=SubcommandA)

    result = tyro.cli(
        tyro.conf.CascadeSubcommandArgs[Config],
        args=[],
        default=Config(),
    )
    assert isinstance(result.subcommand, SubcommandA)
    assert isinstance(result.subcommand.nested, NestedA)



from typing import Generic, TypeVar

from helptext_utils import get_helptext_with_checks
from pydantic import BaseModel

import tyro

T = TypeVar("T")


def test_pydantic_generic() -> None:
    class ManyTypesA(BaseModel, Generic[T]):
        i: T
        s: str = "hello"
        f: float = 3.0

    assert tyro.cli(ManyTypesA[int], args=["--i", "5"]) == ManyTypesA(
        i=5, s="hello", f=3.0
    )


def test_pydantic_inheritance_with_same_typevar() -> None:
    T = TypeVar("T")

    class A(BaseModel, Generic[T]):
        x: T

    class B(A[int], Generic[T]):
        y: T

    assert "INT" in get_helptext_with_checks(B[int])
    assert "STR" not in get_helptext_with_checks(B[int])
    assert "STR" in get_helptext_with_checks(B[str])
    assert "INT" in get_helptext_with_checks(B[str])

    assert tyro.cli(B[str], args=["--x", "1", "--y", "2"]) == B(x=1, y="2")
    assert tyro.cli(B[int], args=["--x", "1", "--y", "2"]) == B(x=1, y=2)


if __name__ == "__main__":
    test_pydantic_generic()
    # test_pydantic_inheritance_with_same_typevar()



"""Core functionality for calling functions with arguments specified by argparse
namespaces."""

from __future__ import annotations

import dataclasses
import itertools
from functools import partial
from typing import Any, Callable, Generic, TypeVar, Union

from typing_extensions import Annotated

from . import _arguments, _parsers, _resolver, _singleton, _strings
from .conf import _confstruct, _markers
from .constructors._primitive_spec import UnsupportedTypeAnnotationError

T = TypeVar("T")


@dataclasses.dataclass(frozen=True)
class DummyWrapper(Generic[T]):
    __tyro_dummy_inner__: Annotated[T, _confstruct.arg(name="")]


@dataclasses.dataclass(frozen=True)
class InstantiationError(Exception):
    """Exception raised when instantiation fail; this typically means that values from
    the CLI are invalid."""

    message: str
    arg: _arguments.ArgumentDefinition | str


def callable_with_args(
    f: Callable[..., T],
    parser_definition: _parsers.ParserSpecification,
    default_instance: T | _singleton.NonpropagatingMissingType,
    value_from_prefixed_field_name: dict[str | None, Any],
    field_name_prefix: str,
) -> tuple[Callable[[], T], set[str]]:
    """Populate `f` with arguments specified by a dictionary of values from argparse.

    Returns a partialed version of `f` with arguments populated, and a set of
    used arguments.

    We return a `Callable[[], OutT]` instead of `T` directly for aesthetic
    reasons; it lets use reduce layers in stack traces for errors from
    functions passed to `tyro`.
    """

    # If we' returning the default: unwrap any dummy wrappers.
    while isinstance(default_instance, DummyWrapper):
        default_instance = default_instance.__tyro_dummy_inner__

    positional_args: list[Any] = []
    kwargs: dict[str, Any] = {}
    consumed_keywords: set[str] = set()

    def get_value_from_arg(
        prefixed_field_name: str, arg: _arguments.ArgumentDefinition
    ) -> tuple[Any, bool]:
        """Helper for getting values from `value_from_arg` + doing some extra
        asserts.

        Returns:
            - The value from `value_from_prefixed_field_name`.
            - If the value was found. If True, we found the value (and it will
              be returned as a string or list of strings). If False, we've just
              returned the default.
        """

        if prefixed_field_name not in value_from_prefixed_field_name:
            # When would the value not be found?
            # 1. If the argument is suppressed
            # 2. If we have `tyro.conf.CascadeSubcommandArgs` at parser level
            # 3. If the argument has the CascadeSubcommandArgs marker
            assert (
                arg.is_suppressed()
                or (_markers.CascadeSubcommandArgs in parser_definition.markers)
                or (_markers.CascadeSubcommandArgs in arg.field.markers)
            ), (
                f"Field value for {arg.lowered.name_or_flags} is unexpectedly missing. This is likely a bug in tyro."
            )
            return arg.field.default, False
        else:
            return value_from_prefixed_field_name[prefixed_field_name], True

    arg_from_prefixed_field_name: dict[str, _arguments.ArgumentDefinition] = {}
    for arg in parser_definition.args:
        arg_from_prefixed_field_name[
            _strings.make_field_name([arg.intern_prefix, arg.field.intern_name])
        ] = arg

    any_arguments_provided = False

    for field in parser_definition.field_list:
        value: Any
        prefixed_field_name = _strings.make_field_name(
            [field_name_prefix, field.intern_name]
        )

        # Resolve field type.
        field_type = field.type_stripped
        if prefixed_field_name in arg_from_prefixed_field_name:
            assert prefixed_field_name not in consumed_keywords

            # Standard arguments.
            arg = arg_from_prefixed_field_name[prefixed_field_name]
            name_maybe_prefixed = prefixed_field_name
            consumed_keywords.add(name_maybe_prefixed)
            if not arg.lowered.is_fixed():
                value, value_found = get_value_from_arg(name_maybe_prefixed, arg)
                should_cast = False

                if _singleton.is_missing(value):
                    value = arg.field.default

                    # Consider a function with a positional sequence argument:
                    #
                    #     def f(x: tuple[int, ...], /)
                    #
                    # If we run this script with no arguments, we should interpret this
                    # as empty input for x. But the argparse default will be a MISSING
                    # value, and the field default will be inspect.Parameter.empty.
                    if (
                        _singleton.is_missing(value)
                        and arg.is_positional()
                        # nargs="?" is currently only used for optional positional
                        # arguments when the underlying nargs for the primitive
                        # constructor is 1. Logic for this is in _arguments.py.
                        and arg.lowered.nargs == "*"
                    ):
                        value = []
                        should_cast = True
                elif value_found:
                    # Value was found from the CLI, so we need to cast it with instance_from_str.
                    should_cast = True
                    any_arguments_provided = True
                    if arg.lowered.nargs == "?":
                        # Special case for optional positional arguments: this is the
                        # only time that arguments don't come back as a list.
                        value = [value]

                # Attempt to cast the value to the correct type.
                if should_cast:
                    try:
                        assert arg.lowered.instance_from_str is not None
                        value = arg.lowered.instance_from_str(value)
                    except (ValueError, TypeError) as e:
                        raise InstantiationError(
                            e.args[0],
                            arg,
                        )
            else:
                assert not _singleton.is_missing(arg.field.default)
                value = arg.field.default
                parsed_value = value_from_prefixed_field_name.get(
                    prefixed_field_name, _singleton.MISSING_NONPROP
                )
                if not _singleton.is_missing(parsed_value):
                    raise InstantiationError(
                        f"{'/'.join(arg.lowered.name_or_flags)} was passed in, but"
                        " is a fixed argument that cannot be parsed",
                        arg,
                    )
        elif prefixed_field_name in parser_definition.child_from_prefix:
            # Nested callable.
            if _resolver.unwrap_origin_strip_extras(field_type) is Union:
                field_type = type(field.default)
            get_value, consumed_keywords_child = callable_with_args(
                field_type,
                parser_definition.child_from_prefix[prefixed_field_name],
                field.default,
                value_from_prefixed_field_name,
                field_name_prefix=prefixed_field_name,
            )
            value = get_value()
            del get_value
            consumed_keywords |= consumed_keywords_child
        else:
            # Unions over dataclasses (subparsers). This is the only other option.
            subparser_def = parser_definition.subparsers_from_intern_prefix[
                prefixed_field_name
            ]
            subparser_dest = _strings.make_subparser_dest(name=prefixed_field_name)
            consumed_keywords.add(subparser_dest)
            if subparser_dest in value_from_prefixed_field_name:
                subparser_name = value_from_prefixed_field_name[subparser_dest]
            else:
                assert not _singleton.is_missing(subparser_def.default_instance), (
                    f"{subparser_dest} missing, but no default instance set. {value_from_prefixed_field_name.keys()=}"
                )
                subparser_name = None

            if subparser_name is None:
                # No subparser selected -- this should only happen when we have a
                # default/default_factory set.
                # This assert is wrong because `type(None)` can be in an `Annotated metadata.
                # assert (
                #     type(None) in get_args(field_type)
                #     or subparser_def.default_instance is not None
                # )
                value = subparser_def.default_instance
            else:
                chosen_f = subparser_def.options[
                    list(subparser_def.parser_from_name.keys()).index(subparser_name)
                ]
                evaluated = subparser_def.parser_from_name[subparser_name].evaluate()
                # Error should have been caught earlier.
                assert not isinstance(evaluated, UnsupportedTypeAnnotationError), (
                    "Unexpected UnsupportedTypeAnnotationError in backend"
                )
                get_value, consumed_keywords_child = callable_with_args(
                    chosen_f,
                    evaluated,
                    (
                        field.default
                        if type(field.default) is chosen_f
                        else _singleton.MISSING_NONPROP
                    ),
                    value_from_prefixed_field_name,
                    field_name_prefix=prefixed_field_name,
                )
                value = get_value()
                del get_value
                consumed_keywords |= consumed_keywords_child

        if value is _singleton.EXCLUDE_FROM_CALL:
            continue

        if field.call_mode == "unpack_args":
            if len(positional_args) == 0 and len(kwargs) > 0:
                positional_args.extend(kwargs.values())
                kwargs.clear()
            # Handle missing value for *args - track it for _OPTIONAL_GROUP logic.
            if _singleton.is_missing(value):
                positional_args.append(value)
            else:
                assert isinstance(value, tuple)
                positional_args.extend(value)
        elif field.call_mode == "unpack_kwargs":
            # Handle missing value for **kwargs - track it for _OPTIONAL_GROUP logic.
            if _singleton.is_missing(value):
                kwargs[field.call_argname] = value
            else:
                assert isinstance(value, dict)
                kwargs.update(value)
        elif field.call_mode == "positional":
            assert len(kwargs) == 0
            positional_args.append(value)
        else:
            kwargs[field.call_argname] = value

    # Logic for _markers._OPTIONAL_GROUP.
    is_missing_list = [
        _singleton.is_missing(v)
        for v in itertools.chain(positional_args, kwargs.values())
    ]
    if any(is_missing_list):
        if not any_arguments_provided:
            # No arguments were provided in this group.
            return lambda: default_instance, consumed_keywords  # type: ignore

        message = "either all arguments must be provided or none of them."
        if len(kwargs) > 0:
            missing_args: list[str] = []
            for k, v in kwargs.items():
                if not _singleton.is_missing(v):
                    break

                # Argument is missing.
                found = False
                for arg in arg_from_prefixed_field_name.values():
                    if arg.field.call_argname == k:
                        missing_args.append("/".join(arg.lowered.name_or_flags))
                        found = True
                        break
                assert found, "This is likely a bug in tyro."

            if len(missing_args) > 0:
                message += f" We're missing arguments {missing_args}."
        raise InstantiationError(
            message,
            field_name_prefix,
        )

    unwrapped_f = parser_definition.f
    if unwrapped_f in (tuple, list, set):
        if len(positional_args) > 0:
            # Triggered when support_single_arg_types=True is used.
            assert len(kwargs) == 0
            assert len(positional_args) == 1
            return lambda: positional_args[0], consumed_keywords  # type: ignore
        else:
            assert len(positional_args) == 0
            return partial(unwrapped_f, kwargs.values()), consumed_keywords  # type: ignore
    elif unwrapped_f is dict:
        if len(positional_args) > 0:
            # Triggered when support_single_arg_types=True is used.
            assert len(kwargs) == 0
            assert len(positional_args) == 1
            return partial(unwrapped_f, *positional_args), consumed_keywords  # type: ignore
        else:
            assert len(positional_args) == 0
            return lambda: kwargs, consumed_keywords  # type: ignore
    else:
        if parser_definition.extern_prefix == "":
            # Don't catch any errors for the "root" field. If main() in
            # tyro.cli(main) raises a ValueError, this shouldn't be caught.
            #
            # Important: we'll unwrap `DummyWrapper` in _cli.py. We could also
            # add an inner function here, but that would make stack traces
            # messier.
            return partial(unwrapped_f, *positional_args, **kwargs), consumed_keywords  # type: ignore
        else:
            # Try to catch ValueErrors raised by field constructors.
            def with_instantiation_error():
                try:
                    out = unwrapped_f(*positional_args, **kwargs)
                # If unwrapped_f raises a ValueError, wrap the message with a more informative
                # InstantiationError if possible.
                except ValueError as e:
                    raise InstantiationError(
                        e.args[0],
                        field_name_prefix,
                    )
                while isinstance(out, DummyWrapper):
                    out = out.__tyro_dummy_inner__
                return out

            return with_instantiation_error, consumed_keywords  # type: ignore



# mypy: ignore-errors
from dataclasses import dataclass
from typing import Any, Generic, Protocol, Type, TypeVar, runtime_checkable

import tyro

OutT_co = TypeVar("OutT_co", covariant=True)


@runtime_checkable
class BuilderProtocol(Protocol[OutT_co]):
    def build(self, **kwd_override: Any) -> OutT_co: ...


OutT = TypeVar("OutT")


class SomeConfig(Generic[OutT]):
    _target: Type[OutT]

    def _configure(self) -> OutT: ...


class Foo: ...


@dataclass
class RealConfig(SomeConfig[Foo], BuilderProtocol[Foo]):
    bar: str = "foo_bar"

    def build(self, **kwd_override: Any) -> Foo:
        return Foo()


def test_protocol_typevar() -> None:
    """Adapted from: https://github.com/brentyi/tyro/issues/335"""
    expected = RealConfig(bar="baz")
    assert tyro.cli(RealConfig, args=["--bar", "baz"]) == expected


test_protocol_typevar()



# mypy: disable-error-code="call-overload,misc"
#
# Mypy errors from passing union types directly into tyro.cli() as Type[T]. We would
# benefit from TypeForm[T]: https://github.com/python/mypy/issues/9773
import collections
import collections.abc
import contextlib
import dataclasses
import enum
import io
import sys
from typing import (
    Any,
    Deque,
    Dict,
    FrozenSet,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
)

import pytest
from helptext_utils import get_helptext_with_checks
from typing_extensions import Literal

import tyro


def test_tuples_fixed() -> None:
    @dataclasses.dataclass
    class A:
        x: Tuple[int, int, int]

    assert tyro.cli(A, args=["--x", "1", "2", "3"]) == A(x=(1, 2, 3))
    with pytest.raises(SystemExit):
        tyro.cli(A, args=["--x"])
    with pytest.raises(SystemExit):
        tyro.cli(A, args=[])


def test_tuples_fixed_mixed() -> None:
    @dataclasses.dataclass
    class A:
        x: Tuple[int, str, int]

    assert tyro.cli(A, args=["--x", "1", "2", "3"]) == A(x=(1, "2", 3))
    with pytest.raises(SystemExit):
        tyro.cli(A, args=["--x"])
    with pytest.raises(SystemExit):
        tyro.cli(A, args=[])


def test_tuples_with_default() -> None:
    @dataclasses.dataclass
    class A:
        x: Tuple[int, int, int] = (0, 1, 2)

    assert tyro.cli(A, args=[]) == A(x=(0, 1, 2))
    assert tyro.cli(A, args=["--x", "1", "2", "3"]) == A(x=(1, 2, 3))
    with pytest.raises(SystemExit):
        tyro.cli(A, args=["--x"])


def test_tuple_with_literal_and_default() -> None:
    @dataclasses.dataclass
    class A:
        x: Tuple[Literal[1, 2, 3], ...] = (1, 2)

    assert tyro.cli(A, args=["--x", "1", "2", "3"]) == A(x=(1, 2, 3))
    assert tyro.cli(A, args=[]) == A(x=(1, 2))
    assert tyro.cli(A, args=["--x"]) == A(x=())
    with pytest.raises(SystemExit):
        tyro.cli(A, args=["--x", "1", "2", "3", "4"])


def test_positional_tuple_with_literal_and_default() -> None:
    @dataclasses.dataclass
    class A:
        x: tyro.conf.Positional[Tuple[Literal[1, 2, 3], ...]] = (1, 2)

    assert tyro.cli(A, args=["1", "2", "3"]) == A(x=(1, 2, 3))
    assert tyro.cli(A, args=[]) == A(x=(1, 2))

    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stderr(target):
        tyro.cli(A, args=["1", "2", "3", "4"])
    assert "invalid choice" in target.getvalue()


def test_tuples_fixed_multitype() -> None:
    @dataclasses.dataclass
    class A:
        x: Tuple[int, str, float]

    assert tyro.cli(A, args=["--x", "1", "2", "3.5"]) == A(x=(1, "2", 3.5))
    with pytest.raises(SystemExit):
        tyro.cli(A, args=["--x"])
    with pytest.raises(SystemExit):
        tyro.cli(A, args=[])


def test_tuples_fixed_bool() -> None:
    @dataclasses.dataclass
    class A:
        x: Tuple[bool, bool, bool]

    assert tyro.cli(A, args=["--x", "True", "True", "False"]) == A(
        x=(True, True, False)
    )
    with pytest.raises(SystemExit):
        tyro.cli(A, args=["--x"])
    with pytest.raises(SystemExit):
        tyro.cli(A, args=[])


def test_tuples_variable() -> None:
    @dataclasses.dataclass
    class A:
        x: Tuple[int, ...]

    assert tyro.cli(A, args=["--x", "1", "2", "3"]) == A(x=(1, 2, 3))
    assert tyro.cli(A, args=["--x"]) == A(x=())
    with pytest.raises(SystemExit):
        tyro.cli(A, args=[])


def test_tuples_variable_bool() -> None:
    @dataclasses.dataclass
    class A:
        x: Tuple[bool, ...]

    assert tyro.cli(A, args=["--x", "True", "True", "False"]) == A(
        x=(True, True, False)
    )
    assert tyro.cli(A, args=["--x"]) == A(x=())
    with pytest.raises(SystemExit):
        tyro.cli(A, args=[])


def test_tuples_variable_optional() -> None:
    @dataclasses.dataclass
    class A:
        x: Optional[Tuple[int, ...]] = None

    assert tyro.cli(A, args=["--x", "1", "2", "3"]) == A(x=(1, 2, 3))
    assert tyro.cli(A, args=["--x"]) == A(x=())
    assert tyro.cli(A, args=[]) == A(x=None)


def test_sequences() -> None:
    @dataclasses.dataclass
    class A:
        x: Sequence[int]

    assert tyro.cli(A, args=["--x", "1", "2", "3"]) == A(x=[1, 2, 3])
    assert tyro.cli(A, args=["--x"]) == A(x=[])
    with pytest.raises(SystemExit):
        tyro.cli(A, args=[])


def test_sequences_narrow() -> None:
    @dataclasses.dataclass
    class A:
        x: Sequence = dataclasses.field(default_factory=lambda: [0])

    assert tyro.cli(A, args=["--x", "1", "2", "3"]) == A(x=[1, 2, 3])
    assert tyro.cli(A, args=[]) == A(x=[0])
    assert tyro.cli(A, args=["--x"]) == A(x=[])


def test_sequences_narrow_any() -> None:
    @dataclasses.dataclass
    class A:
        x: Sequence[Any] = dataclasses.field(default_factory=lambda: [0])

    assert tyro.cli(A, args=["--x", "1", "2", "3"]) == A(x=[1, 2, 3])
    assert tyro.cli(A, args=[]) == A(x=[0])
    assert tyro.cli(A, args=["--x"]) == A(x=[])


if sys.version_info >= (3, 9):

    def test_abc_sequences() -> None:
        @dataclasses.dataclass
        class A:
            x: collections.abc.Sequence[int]

        assert tyro.cli(A, args=["--x", "1", "2", "3"]) == A(x=[1, 2, 3])
        assert tyro.cli(A, args=["--x"]) == A(x=[])
        with pytest.raises(SystemExit):
            tyro.cli(A, args=[])


def test_abc_sequences_narrow() -> None:
    @dataclasses.dataclass
    class A:
        x: collections.abc.Sequence = dataclasses.field(default_factory=lambda: [0])

    assert tyro.cli(A, args=["--x", "1", "2", "3"]) == A(x=[1, 2, 3])
    assert tyro.cli(A, args=[]) == A(x=[0])
    assert tyro.cli(A, args=["--x"]) == A(x=[])


if sys.version_info >= (3, 9):

    def test_abc_sequences_narrow_any() -> None:
        @dataclasses.dataclass
        class A:
            x: collections.abc.Sequence[Any] = dataclasses.field(
                default_factory=lambda: [0]
            )

        assert tyro.cli(A, args=["--x", "1", "2", "3"]) == A(x=[1, 2, 3])
        assert tyro.cli(A, args=[]) == A(x=[0])
        assert tyro.cli(A, args=["--x"]) == A(x=[])


def test_lists() -> None:
    @dataclasses.dataclass
    class A:
        x: List[int]

    assert tyro.cli(A, args=["--x", "1", "2", "3"]) == A(x=[1, 2, 3])
    assert tyro.cli(A, args=["--x"]) == A(x=[])
    with pytest.raises(SystemExit):
        tyro.cli(A, args=[])


def test_list_with_literal() -> None:
    @dataclasses.dataclass
    class A:
        x: List[Literal[1, 2, 3]]

    assert tyro.cli(A, args=["--x", "1", "2", "3"]) == A(x=[1, 2, 3])
    assert tyro.cli(A, args=["--x"]) == A(x=[])
    with pytest.raises(SystemExit):
        tyro.cli(A, args=["--x", "1", "2", "3", "4"])
    with pytest.raises(SystemExit):
        tyro.cli(A, args=[])


def test_list_with_enums() -> None:
    class Color(enum.Enum):
        RED = enum.auto()
        GREEN = enum.auto()
        BLUE = enum.auto()

    @dataclasses.dataclass
    class A:
        x: List[Color]

    assert tyro.cli(A, args=["--x", "RED", "RED", "BLUE"]) == A(
        x=[Color.RED, Color.RED, Color.BLUE]
    )
    assert tyro.cli(A, args=["--x"]) == A(x=[])
    with pytest.raises(SystemExit):
        tyro.cli(A, args=["--x", "RED", "RED", "YELLOW"])
    with pytest.raises(SystemExit):
        tyro.cli(A, args=[])


def test_lists_with_default() -> None:
    class Color(enum.Enum):
        RED = enum.auto()
        GREEN = enum.auto()
        BLUE = enum.auto()

    @dataclasses.dataclass
    class A:
        x: List[Color] = dataclasses.field(
            default_factory=[Color.RED, Color.GREEN].copy
        )

    assert tyro.cli(A, args=[]) == A(x=[Color.RED, Color.GREEN])
    assert tyro.cli(A, args=["--x", "RED", "GREEN", "BLUE"]) == A(
        x=[Color.RED, Color.GREEN, Color.BLUE]
    )


def test_lists_bool() -> None:
    @dataclasses.dataclass
    class A:
        x: List[bool]

    assert tyro.cli(A, args=["--x", "True", "False", "True"]) == A(
        x=[True, False, True]
    )
    assert tyro.cli(A, args=["--x"]) == A(x=[])
    with pytest.raises(SystemExit):
        tyro.cli(A, args=[])


def test_sets() -> None:
    @dataclasses.dataclass
    class A:
        x: Set[int]

    assert tyro.cli(A, args=["--x", "1", "2", "3", "3"]) == A(x={1, 2, 3})
    assert tyro.cli(A, args=["--x"]) == A(set())
    with pytest.raises(SystemExit):
        tyro.cli(A, args=[])


def test_frozen_sets() -> None:
    @dataclasses.dataclass
    class A:
        x: FrozenSet[int]

    assert tyro.cli(A, args=["--x", "1", "2", "3", "3"]) == A(x=frozenset({1, 2, 3}))
    assert tyro.cli(A, args=["--x"]) == A(x=frozenset())
    with pytest.raises(SystemExit):
        tyro.cli(A, args=[])


def test_deque() -> None:
    @dataclasses.dataclass
    class A:
        x: Deque[int]

    assert tyro.cli(A, args=["--x", "1", "2", "3", "3"]) == A(
        x=collections.deque([1, 2, 3, 3])
    )
    assert tyro.cli(A, args=["--x"]) == A(collections.deque())
    with pytest.raises(SystemExit):
        tyro.cli(A, args=[])


def test_sets_with_default() -> None:
    @dataclasses.dataclass
    class A:
        x: Set[int] = dataclasses.field(default_factory={0, 1, 2}.copy)

    assert tyro.cli(A, args=[]) == A(x={0, 1, 2})
    assert tyro.cli(A, args=["--x", "1", "2", "3", "3"]) == A(x={1, 2, 3})
    assert tyro.cli(A, args=["--x"]) == A(x=set())


def test_optional_sequences() -> None:
    @dataclasses.dataclass
    class A:
        x: Optional[Sequence[int]] = None

    assert tyro.cli(A, args=["--x", "1", "2", "3"]) == A(x=[1, 2, 3])
    assert tyro.cli(A, args=["--x"]) == A(x=[])
    assert tyro.cli(A, args=["--x", "None"]) == A(x=None)
    assert tyro.cli(A, args=[]) == A(x=None)


def test_optional_lists() -> None:
    @dataclasses.dataclass
    class A:
        x: Optional[List[int]] = None

    assert tyro.cli(A, args=["--x", "1", "2", "3"]) == A(x=[1, 2, 3])
    assert tyro.cli(A, args=["--x"]) == A(x=[])
    assert tyro.cli(A, args=["--x", "None"]) == A(x=None)
    assert tyro.cli(A, args=[]) == A(x=None)


def test_nested_optional_types() -> None:
    """We support "None" as a special-case keyword. (note: this is a bit weird because
    Optional[str] might interpret "None" as either a string or an actual `None`
    value)"""

    @dataclasses.dataclass
    class A:
        x: Tuple[Optional[int], ...]

    assert tyro.cli(A, args=["--x", "0", "1"]) == A((0, 1))
    assert tyro.cli(A, args=["--x", "0", "None", "1"]) == A((0, None, 1))


def test_union_over_collections() -> None:
    def main(a: Union[Tuple[float, ...], Tuple[int, ...]]) -> Any:
        return a

    assert tyro.cli(main, args="--a 3.3 3.3 7.0".split(" ")) == (3.3, 3.3, 7.0)
    assert tyro.cli(main, args="--a 3 3 7".split(" ")) == (3, 3, 7)


def test_union_over_collections_2() -> None:
    def main(a: Union[Tuple[str, float, str], Tuple[str, str, float]]) -> Any:
        return a

    assert tyro.cli(main, args="--a 3.3 hey 7.0".split(" ")) == ("3.3", "hey", 7.0)
    assert tyro.cli(main, args="--a 3 3 hey".split(" ")) == ("3", 3.0, "hey")


def test_union_over_collections_3() -> None:
    def main(a: Union[Tuple[int, int], Tuple[int, int, int]]) -> Tuple[int, ...]:
        return a

    assert tyro.cli(main, args=["--a", "5", "5"]) == (5, 5)
    assert tyro.cli(main, args=["--a", "1", "2", "3"]) == (1, 2, 3)

    with pytest.raises(SystemExit):
        tyro.cli(main, args=["--a", "5", "5", "2", "1"])

    with pytest.raises(SystemExit):
        tyro.cli(main, args=["--a"])
    with pytest.raises(SystemExit):
        tyro.cli(main, args=[])


def test_choices_in_tuples_0() -> None:
    @dataclasses.dataclass
    class A:
        x: Tuple[bool, bool]

    assert tyro.cli(A, args=["--x", "True", "False"]) == A((True, False))


def test_choices_in_tuples_1() -> None:
    @dataclasses.dataclass
    class A:
        x: Tuple[bool, Literal["True", "False"]]

    assert tyro.cli(A, args=["--x", "True", "False"]) == A((True, "False"))


def test_choices_in_tuples_2() -> None:
    @dataclasses.dataclass
    class A:
        x: Tuple[bool, Literal["True", "False", "None"]]

    assert tyro.cli(A, args=["--x", "True", "False"]).x == (True, "False")
    assert tyro.cli(A, args=["--x", "False", "None"]).x == (False, "None")
    with pytest.raises(SystemExit):
        tyro.cli(A, args=["--x", "None", "False"])


def test_nested_tuple_types() -> None:
    @dataclasses.dataclass
    class A:
        x: Tuple[Tuple[int, int], Tuple[str, str]]

    assert tyro.cli(A, args="--x 5 5 5 5".split(" ")).x == ((5, 5), ("5", "5"))


def test_variable_nested_tuple() -> None:
    def main(x: Tuple[Tuple[int, str], ...]) -> tuple:
        return x

    assert tyro.cli(main, args="--x 1 1 2 2".split(" ")) == ((1, "1"), (2, "2"))
    with pytest.raises(SystemExit):
        tyro.cli(main, args="--x 1 1 2".split(" "))


def test_super_nested() -> None:
    def main(
        x: Optional[
            List[
                Tuple[
                    Optional[int],
                    Literal[3, 4],
                    Union[Tuple[int, int], Tuple[str, str]],
                ]
            ]
        ] = None,
    ) -> Any:
        return x

    assert tyro.cli(main, args=[]) is None
    assert tyro.cli(main, args="--x None".split(" ")) is None
    assert tyro.cli(main, args="--x None 3 2 2".split(" ")) == [(None, 3, (2, 2))]
    assert tyro.cli(main, args="--x 2 3 x 2".split(" ")) == [(2, 3, ("x", "2"))]
    assert tyro.cli(main, args="--x 2 3 x 2 2 3 1 2".split(" ")) == [
        (2, 3, ("x", "2")),
        (2, 3, (1, 2)),
    ]
    with pytest.raises(SystemExit):
        tyro.cli(main, args=["--help"])


def test_dict_no_annotation() -> None:
    def main(x: Dict[str, Any] = {"int": 5, "str": "5"}):
        return x

    assert tyro.cli(main, args=[]) == {"int": 5, "str": "5"}
    assert tyro.cli(main, args="--x.int 3 --x.str 7".split(" ")) == {
        "int": 3,
        "str": "7",
    }


def test_dict_no_annotation_2() -> None:
    def main(x: Dict = {"int": 5, "str": "5"}):
        return x

    assert tyro.cli(main, args=[]) == {"int": 5, "str": "5"}
    assert tyro.cli(main, args="--x.int 3 --x.str 7".split(" ")) == {
        "int": 3,
        "str": "7",
    }


def test_dict_optional() -> None:
    # In this case, the `None` is ignored.
    def main(x: Optional[Dict[str, float]] = {"three": 3, "five": 5}):
        return x

    assert tyro.cli(main, args=[]) == {"three": 3, "five": 5}
    assert tyro.cli(main, args="--x 3 3 5 5".split(" ")) == {"3": 3, "5": 5}


def test_double_dict_no_annotation() -> None:
    def main(
        x: Dict[str, Any] = {
            "wow": {"int": 5, "str": "5"},
        },
    ):
        return x

    assert tyro.cli(main, args=[]) == {"wow": {"int": 5, "str": "5"}}
    assert tyro.cli(main, args="--x.wow.int 3 --x.wow.str 7".split(" ")) == {
        "wow": {
            "int": 3,
            "str": "7",
        }
    }


def test_list_narrowing() -> None:
    def main(x: list = [0, 1, 2, "hello"]) -> Any:
        return x

    assert tyro.cli(main, args="--x hi there 5".split(" ")) == ["hi", "there", 5]


def test_list_narrowing_any() -> None:
    def main(x: List[Any] = [0, 1, 2, "hello"]) -> Any:
        return x

    assert tyro.cli(main, args="--x hi there 5".split(" ")) == ["hi", "there", 5]


def test_list_narrowing_empty() -> None:
    def main(x: list = []) -> Any:
        return x

    assert tyro.cli(main, args="--x hi there 5".split(" ")) == ["hi", "there", "5"]


def test_list_narrowing_empty_any() -> None:
    def main(x: List[Any] = []) -> Any:
        return x

    assert tyro.cli(main, args="--x hi there 5".split(" ")) == ["hi", "there", "5"]


def test_set_narrowing() -> None:
    def main(x: set = {0, 1, 2, "hello"}) -> Any:
        return x

    out = tyro.cli(main, args="--x hi there 5".split(" "))
    # Nondeterministic depending on set iteration order, `int | str` or `str | int`.
    assert out in ({"hi", "there", 5}, {"hi", "there", "5"})


def test_set_narrowing_any() -> None:
    def main(x: Set[Any] = {0, 1, 2, "hello"}) -> Any:
        return x

    out = tyro.cli(main, args="--x hi there 5".split(" "))
    # Nondeterministic depending on set iteration order, `int | str` or `str | int`.
    assert out in ({"hi", "there", 5}, {"hi", "there", "5"})


def test_set_narrowing_empty() -> None:
    def main(x: set = set()) -> Any:
        return x

    assert tyro.cli(main, args="--x hi there 5".split(" ")) == {"hi", "there", "5"}


def test_set_narrowing_any_empty() -> None:
    def main(x: Set[Any] = set()) -> Any:
        return x

    assert tyro.cli(main, args="--x hi there 5".split(" ")) == {"hi", "there", "5"}


def test_tuple_narrowing() -> None:
    def main(x: tuple = (0, 1, 2, "hello")) -> Any:
        return x

    assert tyro.cli(main, args="--x 0 1 2 3".split(" ")) == (0, 1, 2, "3")


def test_tuple_narrowing_any() -> None:
    def main(x: Tuple[Any, ...] = (0, 1, 2, "hello")) -> Any:
        return x

    assert tyro.cli(main, args="--x 0 1 2 3".split(" ")) == (0, 1, 2, "3")


def test_tuple_narrowing_empty() -> None:
    def main(x: tuple = ()) -> Any:
        return x

    assert tyro.cli(main, args="--x 0 1 2 3".split(" ")) == ("0", "1", "2", "3")


def test_tuple_narrowing_empty_any() -> None:
    def main(x: Tuple[Any, ...] = ()) -> Any:
        return x

    assert tyro.cli(main, args="--x 0 1 2 3".split(" ")) == ("0", "1", "2", "3")


def test_tuple_narrowing_empty_default() -> None:
    def main(x: tuple = ()) -> Any:
        return x

    assert tyro.cli(main, args="--x 0 1 2 3".split(" ")) == ("0", "1", "2", "3")


def test_narrowing_edge_case() -> None:
    """https://github.com/brentyi/tyro/issues/136"""

    @dataclasses.dataclass
    class Config:
        _target: Type = dataclasses.field(default_factory=lambda: MyClass)

    class MyClass:
        def __len__(self):
            return 0

    assert tyro.cli(Config, args=[]) == Config()


def test_no_type_collections():
    assert tyro.cli(dict, args="a b c d".split(" ")) == {"a": "b", "c": "d"}
    assert tyro.cli(list, args="a b c d".split(" ")) == ["a", "b", "c", "d"]
    assert tyro.cli(tuple, args="a b c d".split(" ")) == ("a", "b", "c", "d")
    assert tyro.cli(set, args="a b c d".split(" ")) == {"a", "b", "c", "d"}


def test_list_narrowing_alt() -> None:
    def main(x: list = [1, "1"]) -> list:
        return x

    assert tyro.cli(main, args="--x 3 four 5".split(" ")) == [3, "four", 5]


def test_list_narrowing_direct() -> None:
    assert tyro.cli(list, default=[1, "2"], args="3 four 5".split(" ")) == [
        3,
        "four",
        5,
    ]


def test_tuple_direct() -> None:
    assert tyro.cli(Tuple[int, ...], args="1 2".split(" ")) == (1, 2)  # type: ignore
    assert tyro.cli(Tuple[int, int], args="1 2".split(" ")) == (1, 2)  # type: ignore


def test_nested_dict_in_list() -> None:
    """https://github.com/nerfstudio-project/nerfstudio/pull/3567"""

    @dataclasses.dataclass
    class Args:
        proposal_net_args_list: List[Dict] = dataclasses.field(
            default_factory=lambda: [
                {
                    "hidden_dim": 16,
                },
                {
                    "hidden_dim": 16,
                },
            ]
        )
        proposal_net_args_list2: Tuple[Dict[str, List], Dict[str, List]] = (
            dataclasses.field(
                default_factory=lambda: (
                    {
                        "hidden_dim": [16, 32],
                    },
                    {
                        "hidden_dim": [16, 32],
                    },
                )
            )
        )

    assert tyro.cli(Args, args=[]) == Args()
    assert tyro.cli(
        Args, args=["--proposal-net-args-list.0.hidden-dim", "32"]
    ).proposal_net_args_list == (
        [
            {
                "hidden_dim": 32,
            },
            {
                "hidden_dim": 16,
            },
        ]
    )
    assert tyro.cli(
        Args, args=["--proposal-net-args-list2.1.hidden-dim", "32", "64"]
    ).proposal_net_args_list2 == (
        {
            "hidden_dim": [16, 32],
        },
        {
            "hidden_dim": [32, 64],
        },
    )


def test_double_list_in_tuple() -> None:
    def main(x: Tuple[List[str], List[str]]) -> None:
        del x

    # This used to be ambiguous, we now break it into two separate arguments!
    helptext = get_helptext_with_checks(main)
    assert "--x.0 [STR [STR ...]]" in helptext
    assert "--x.1 [STR [STR ...]]" in helptext


def test_ambiguous_collection_0() -> None:
    @dataclasses.dataclass
    class A:
        x: Tuple[Tuple[int, ...], ...]

    assert tyro.cli(A, args=["--x", "0", "1"]) == A(x=((0, 1),))


def test_ambiguous_collection_1() -> None:
    @dataclasses.dataclass
    class A:
        x: List[List[int]]

    assert tyro.cli(A, args=["--x", "0", "1"]).x == [[0, 1]]


def test_ambiguous_collection_2() -> None:
    assert tyro.cli(Tuple[List[str], ...], args=["a", "b", "c", "d"]) == (
        ["a", "b", "c", "d"],
    )


def test_ambiguous_collection_3() -> None:
    assert tyro.cli(Dict[Tuple[int, ...], str], args=["0", "1", "a"]) == {(0, 1): "a"}


def test_ambiguous_collection_4() -> None:
    @dataclasses.dataclass
    class A:
        x: Dict[str, List[int]]

    assert tyro.cli(A, args=["--x", "a", "0", "1"]).x == {"a": [0, 1]}


def test_ambiguous_collection_5() -> None:
    @dataclasses.dataclass
    class A:
        x: Dict[Tuple[int, ...], str]

    assert tyro.cli(A, args=["--x", "0", "1", "a"]).x == {(0, 1): "a"}


def test_ambiguous_collection_6() -> None:
    @dataclasses.dataclass
    class A:
        x: Dict[Tuple[Tuple[Tuple[int, ...]]], str]

    assert tyro.cli(A, args=["--x", "0", "1", "a", "2", "3"]).x == {
        (((0, 1),),): "a",
        (((2,),),): "3",
    }


def test_ambiguous_collection_7() -> None:
    @dataclasses.dataclass
    class A:
        x: Tuple[List[str], int, List[str]]

    # We resolve ambiguity in fixed-length tuples by breaking inputs into
    # separate arguments.
    assert tyro.cli(
        A, args=["--x.0", "0", "1", "a", "2", "3", "--x.1", "0", "--x.2", "4"]
    ).x == (
        ["0", "1", "a", "2", "3"],
        0,
        ["4"],
    )



import dataclasses
from typing import Union

import pytest
from helptext_utils import get_helptext_with_checks

import tyro


def test_flag_default_false() -> None:
    """Test for argparse.BooleanOptionalAction-style usage."""

    @dataclasses.dataclass
    class A:
        x: bool

    assert tyro.cli(
        A,
        args=["--x"],
        default=A(False),
    ) == A(True)

    assert tyro.cli(
        A,
        args=["--no-x"],
        default=A(False),
    ) == A(False)

    assert tyro.cli(
        A,
        args=[],
        default=A(False),
    ) == A(False)

    # Type ignore can be removed once TypeForm lands.
    # https://discuss.python.org/t/typeform-spelling-for-a-type-annotation-object-at-runtime/51435
    assert tyro.cli(
        tyro.conf.FlagConversionOff[A],
        args=["--x", "True"],
        default=A(False),  # type: ignore
    ) == A(True)


def test_flag_default_true() -> None:
    """Test for argparse.BooleanOptionalAction-style usage."""

    @dataclasses.dataclass
    class A:
        x: bool

    assert tyro.cli(
        A,
        args=["--x"],
        default=A(True),
    ) == A(True)

    assert tyro.cli(
        A,
        args=["--no-x"],
        default=A(True),
    ) == A(False)

    assert tyro.cli(
        A,
        args=[],
        default=A(True),
    ) == A(True)

    # Type ignore can be removed once TypeForm lands.
    # https://discuss.python.org/t/typeform-spelling-for-a-type-annotation-object-at-runtime/51435
    assert tyro.cli(
        tyro.conf.FlagConversionOff[A],
        args=["--x", "True"],
        default=A(False),  # type: ignore
    ) == A(True)


def test_flag_default_true_helptext() -> None:
    """Test for argparse.BooleanOptionalAction-style usage."""

    @dataclasses.dataclass
    class A:
        x: bool = True

    assert "(default: True)" in get_helptext_with_checks(A)
    assert "(default: False)" not in get_helptext_with_checks(A)
    assert "(default: None)" not in get_helptext_with_checks(A)


def test_flag_no_pairs() -> None:
    """Test for tyro.conf.FlagPairOff."""

    @dataclasses.dataclass
    class A:
        x: tyro.conf.FlagCreatePairsOff[bool]
        y: tyro.conf.FlagCreatePairsOff[bool] = False
        z: tyro.conf.FlagCreatePairsOff[bool] = True

    assert tyro.cli(
        A,
        args=["--x", "True"],
    ) == A(True)
    assert tyro.cli(
        A,
        args=["--x", "True", "--y"],
    ) == A(True, True)
    assert tyro.cli(
        A,
        args=["--x", "True", "--y", "--no-z"],
    ) == A(True, True, False)

    with pytest.raises(SystemExit):
        tyro.cli(
            A,
            args=["--x", "True", "--y", "True"],
        )
    with pytest.raises(SystemExit):
        tyro.cli(
            A,
            args=["--x", "True", "--no-y"],
        )
    with pytest.raises(SystemExit):
        tyro.cli(
            A,
            args=["--x", "True", "--z"],
        )
    with pytest.raises(SystemExit):
        tyro.cli(
            A,
            args=["--x"],
        )
    with pytest.raises(SystemExit):
        tyro.cli(
            A,
            args=["--no-x"],
        )


def test_disallow_none_bool_flag() -> None:
    """Test for DisallowNone[bool | None] flag syntax (issue #381)."""

    @dataclasses.dataclass
    class A:
        x: tyro.conf.DisallowNone[Union[bool, None]] = None

    # Test that flag syntax works.
    assert tyro.cli(A, args=["--x"]) == A(True)
    assert tyro.cli(A, args=["--no-x"]) == A(False)
    assert tyro.cli(A, args=[]) == A(None)

    # Test with different default values.
    @dataclasses.dataclass
    class B:
        x: tyro.conf.DisallowNone[Union[bool, None]] = False

    assert tyro.cli(B, args=["--x"]) == B(True)
    assert tyro.cli(B, args=["--no-x"]) == B(False)
    assert tyro.cli(B, args=[]) == B(False)

    @dataclasses.dataclass
    class C:
        x: tyro.conf.DisallowNone[Union[bool, None]] = True

    assert tyro.cli(C, args=["--x"]) == C(True)
    assert tyro.cli(C, args=["--no-x"]) == C(False)
    assert tyro.cli(C, args=[]) == C(True)


def test_disallow_none_bool_helptext() -> None:
    """Test that DisallowNone[bool | None] shows flag-style helptext."""

    @dataclasses.dataclass
    class A:
        x: tyro.conf.DisallowNone[Union[bool, None]] = None

    helptext = get_helptext_with_checks(A)
    # Should show --x | --no-x instead of {True,False}.
    assert "--x" in helptext or "--x," in helptext
    assert "--no-x" in helptext
    assert "{True,False}" not in helptext
    assert "(default: None)" in helptext


def test_disallow_none_non_union_bool() -> None:
    """Test that DisallowNone on a plain bool (not a union) still works normally."""

    @dataclasses.dataclass
    class A:
        x: tyro.conf.DisallowNone[bool] = False

    # Should still create flag syntax.
    assert tyro.cli(A, args=["--x"]) == A(True)
    assert tyro.cli(A, args=["--no-x"]) == A(False)
    assert tyro.cli(A, args=[]) == A(False)


def test_disallow_none_non_union_int() -> None:
    """Test that DisallowNone on a plain int (not a union) doesn't get flag treatment."""

    @dataclasses.dataclass
    class A:
        x: tyro.conf.DisallowNone[int] = 5

    # Should not create flag syntax, just regular int argument.
    assert tyro.cli(A, args=["--x", "10"]) == A(10)
    assert tyro.cli(A, args=[]) == A(5)

    helptext = get_helptext_with_checks(A)
    # Should not have --no-x since it's not a boolean.
    assert "--no-x" not in helptext


def test_disallow_none_multi_type_union() -> None:
    """Test that DisallowNone[int | str | None] doesn't get treated as a bool."""

    @dataclasses.dataclass
    class A:
        x: tyro.conf.DisallowNone[Union[int, str, None]] = None

    # Should not create flag syntax, should use choice-based syntax.
    helptext = get_helptext_with_checks(A)
    assert "--x" in helptext
    # Should not have --no-x since it's not a boolean.
    assert "--no-x" not in helptext



"""Test TypeVar default support from Python 3.13's typing module.

This test file requires Python 3.13+ since it tests the native typing.TypeVar
default support introduced in PEP 696.
"""

from __future__ import annotations

import typing
import warnings
from dataclasses import dataclass
from typing import Generic

import pytest
from typing_extensions import TypeVar as ExtTypeVar

import tyro
from tyro._warnings import TyroWarning

T_ext = ExtTypeVar("T_ext", default=int)
T_typing = typing.TypeVar("T_typing", default=int)

# Python 3.13+ typing.TypeVar with default.
T_with_default = typing.TypeVar("T_with_default", default=str)
T_no_default = typing.TypeVar("T_no_default")
T_with_bound = typing.TypeVar("T_with_bound", bound=int)


@dataclass
class GenericWithDefault(Generic[T_with_default]):
    """Generic class with typing.TypeVar that has a default."""

    value: T_with_default


@dataclass
class GenericNoDefault(Generic[T_no_default]):
    """Generic class with typing.TypeVar that has no default."""

    value: T_no_default


@dataclass
class GenericWithBound(Generic[T_with_bound]):
    """Generic class with typing.TypeVar that has a bound."""

    value: T_with_bound


def test_typing_typevar_with_default_no_warning():
    """Test that typing.TypeVar with default (Python 3.13+) doesn't produce warning."""
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")

        @dataclass
        class Config(Generic[T_with_default]):
            value: T_with_default = "default"  # type: ignore

        # This should not produce a warning in Python 3.13+.
        result = tyro.cli(Config, args=["--value", "test"], default=Config())

        # Check no TyroWarning was raised for TypeVar resolution.
        tyro_warnings = [
            warning
            for warning in w
            if issubclass(warning.category, TyroWarning)
            and "Could not resolve type parameter" in str(warning.message)
        ]
        assert len(tyro_warnings) == 0, f"Unexpected TypeVar warnings: {tyro_warnings}"
        assert result.value == "test"


def test_typing_typevar_no_default_produces_warning():
    """Test that typing.TypeVar without default produces TyroWarning."""
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")

        @dataclass
        class Config(Generic[T_no_default]):
            value: T_no_default = "default"  # type: ignore

        # This should produce a warning.
        result = tyro.cli(Config, args=["--value", "test"], default=Config())

        # Check that a TyroWarning was raised.
        tyro_warnings = [
            warning
            for warning in w
            if issubclass(warning.category, TyroWarning)
            and "Could not resolve type parameter" in str(warning.message)
        ]
        assert len(tyro_warnings) > 0, "Expected TyroWarning for unresolved TypeVar"
        assert result.value == "test"


def test_typing_typevar_with_bound_produces_warning():
    """Test that typing.TypeVar with bound produces TyroWarning but uses bound."""
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")

        @dataclass
        class Config(Generic[T_with_bound]):
            value: T_with_bound = 42  # type: ignore

        # This should produce a warning but use the bound type.
        result = tyro.cli(Config, args=["--value", "100"], default=Config())

        # Check that a TyroWarning was raised.
        tyro_warnings = [
            warning
            for warning in w
            if issubclass(warning.category, TyroWarning)
            and "Could not resolve type parameter" in str(warning.message)
        ]
        assert len(tyro_warnings) > 0, "Expected TyroWarning for TypeVar with bound"
        assert result.value == 100


def test_both_typing_and_typing_extensions_work():
    """Verify that both typing.TypeVar and typing_extensions.TypeVar work with defaults."""

    @dataclass
    class ConfigExt(Generic[T_ext]):
        value: T_ext = 42  # type: ignore

    @dataclass
    class ConfigTyping(Generic[T_typing]):
        value: T_typing = 42  # type: ignore

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")

        # Both should work without warnings.
        result_ext = tyro.cli(ConfigExt, args=["--value", "100"], default=ConfigExt())
        result_typing = tyro.cli(
            ConfigTyping, args=["--value", "200"], default=ConfigTyping()
        )

        # Check no TypeVar warnings were raised.
        tyro_warnings = [
            warning
            for warning in w
            if issubclass(warning.category, TyroWarning)
            and "Could not resolve type parameter" in str(warning.message)
        ]
        assert len(tyro_warnings) == 0, f"Unexpected TypeVar warnings: {tyro_warnings}"
        assert result_ext.value == 100
        assert result_typing.value == 200


if __name__ == "__main__":
    pytest.main([__file__, "-v"])
